__all__ = ["DomoInstanceConfig"]


from dataclasses import dataclass, field
from typing import Any

import httpx
import pandas as pd

from ...auth import DomoAuth
from ...base.exceptions import ClassError
from ...routes import (
    application as application_routes,
    sandbox as sandbox_routes,
)
from ...routes.auth import InvalidAuthTypeError
from ...routes.instance_config import (
    authorized_domains as domains_routes,
    toggle as toggle_routes,
)
from .access_token import DomoAccessTokens
from .allowlist import DomoAllowlist
from .api_client import ApiClients
from .bootstrap import DomoBootstrap
from .instance_switcher import InstanceSwitcher
from .mfa import MFA_Config
from .role import DomoRoles
from .role_grant import DomoGrants
from .sso import SSO as SSO_Class
from .toggle import DomoToggle
from .user_attributes import UserAttributes


@dataclass
class DomoInstanceConfig:
    """utility class that absorbs many of the domo instance configuration methods"""

    auth: DomoAuth = field(repr=False)

    Accounts: Any = field(default=None)
    AccessTokens: DomoAccessTokens = field(default=None)
    Allowlist: DomoAllowlist = field(default=None)
    ApiClients: "ApiClients" = field(default=None)
    Bootstrap: DomoBootstrap = field(default=None)

    Connectors: Any = field(default=None)  # DomoConnectors
    InstanceSwitcher: "InstanceSwitcher" = field(default=None)

    Grants: DomoGrants = field(default=None)

    MFA: MFA_Config = field(default=None)
    Roles: DomoRoles = field(default=None)

    SSO: SSO_Class = field(default=None)
    Everywhere: Any = field(
        default=None
    )  # DomoEverywhere - imported lazily to avoid circular import
    UserAttributes: "UserAttributes" = field(default=None)
    Toggle: DomoToggle = field(default=None)

    def __post_init__(self):
        from ..DomoAccount import DomoAccounts
        from ..DomoDataset.connector import DomoConnectors
        from ..DomoEverywhere import DomoEverywhere
        from .api_client import ApiClients
        from .instance_switcher import InstanceSwitcher
        from .user_attributes import UserAttributes

        self.Accounts = DomoAccounts(auth=self.auth)
        self.AccessTokens = DomoAccessTokens(auth=self.auth)
        self.ApiClients = ApiClients(auth=self.auth)
        self.Allowlist = DomoAllowlist(auth=self.auth)
        self.Bootstrap = DomoBootstrap(auth=self.auth)

        self.Connectors = DomoConnectors(auth=self.auth)
        self.Grants = DomoGrants(auth=self.auth)
        self.InstanceSwitcher = InstanceSwitcher(auth=self.auth)
        self.MFA = MFA_Config(auth=self.auth)
        self.Everywhere = DomoEverywhere(auth=self.auth)
        self.UserAttributes = UserAttributes(auth=self.auth)
        self.Roles = DomoRoles(auth=self.auth)
        self.SSO = SSO_Class(auth=self.auth)

    async def get_applications(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        from ..DomoApplication.Application import DomoApplication

        res = await application_routes.get_applications(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return [DomoApplication.from_dict(job, auth=self.auth) for job in res.response]

    async def generate_applications_report(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        domo_apps = await self.get_applications(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

        if return_raw:
            return domo_apps

        df = pd.DataFrame([app.__dict__ for app in domo_apps])
        df["domo_instance"] = self.auth.domo_instance

        df.drop(columns=["auth"], inplace=True)
        df.rename(
            columns={
                "id": "application_id",
                "name": "application_name",
                "description": "application_description",
                "version": "application_version",
            },
            inplace=True,
        )

        return df.sort_index(axis=1)

    async def get_authorized_domains(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
    ) -> list[str]:
        """returns a list of authorized domains (str) does not update instance_config"""

        res = await domains_routes.get_authorized_domains(
            auth=self.auth, debug_api=debug_api, session=session, return_raw=return_raw
        )

        if return_raw:
            return res

        return res.response

    async def set_authorized_domains(
        self,
        authorized_domains: list[str],
        debug_api: bool = False,
        debug_num_stacks_to_drop=1,
        session: httpx.AsyncClient = None,
    ):
        res = await domains_routes.set_authorized_domains(
            auth=self.auth,
            authorized_domain_ls=authorized_domains,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        return res

    async def upsert_authorized_domains(
        self,
        authorized_domains: list[str],
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        existing_domains = await self.get_authorized_domains(
            debug_api=debug_api,
            session=session,
        )

        authorized_domains += existing_domains

        return await self.set_authorized_domains(
            authorized_domains=authorized_domains,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

    async def get_authorized_custom_app_domains(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ) -> list[str]:
        res = await domains_routes.get_authorized_custom_app_domains(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            return_raw=return_raw,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return res.response

        # | exporti

    async def set_authorized_custom_app_domains(
        self,
        authorized_domains: list[str],
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        res = await domains_routes.set_authorized_custom_app_domains(
            auth=self.auth,
            authorized_custom_app_domain_ls=authorized_domains,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=self.__class__.__name__,
        )

        return res

    async def upsert_authorized_custom_app_domains(
        self,
        authorized_domains: list[str],
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        existing_domains = await self.get_authorized_custom_app_domains(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        authorized_domains += existing_domains

        return await self.set_authorized_custom_app_domains(
            authorized_domains=authorized_domains,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            session=session,
        )

    async def get_sandbox_is_same_instance_promotion_enabled(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        res = await sandbox_routes.get_is_allow_same_instance_promotion_enabled(
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        self.is_sandbox_self_instance_promotion_enabled = res.response["is_enabled"]

        if return_raw:
            return res

        return res.response

    async def toggle_sandbox_allow_same_instance_promotion(
        self,
        is_enabled: bool,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        """will enable or disable same instance promotion for sandbox"""

        res = await sandbox_routes.toggle_allow_same_instance_promotion(
            auth=self.auth,
            session=session,
            is_enabled=is_enabled,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        res_is_enabled = await self.get_sandbox_is_same_instance_promotion_enabled()

        if return_raw:
            return res

        return res_is_enabled

    async def get_is_user_invite_notification_enabled(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        """
        Admin > Company Settings > Admin Notifications
        Toggles whether user recieves 'You've been Domo'ed email
        """

        res = await toggle_routes.get_is_user_invite_notifications_enabled(
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        self.is_user_invite_notification_enabled = res.response["is_enabled"]

        if return_raw:
            return res

        return res.response

    async def toggle_is_user_invite_notification_enabled(
        self,
        is_enabled: bool,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        res_is_enabled = await self.get_is_user_invite_notification_enabled()

        if is_enabled == self.is_user_invite_notification_enabled:
            return res_is_enabled

        res = await toggle_routes.toggle_is_user_invite_enabled(
            auth=self.auth,
            is_enabled=is_enabled,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        res_is_enabled = await self.get_is_user_invite_notification_enabled()

        if return_raw:
            return res

        return res_is_enabled

    class InstanceConfig_ClassError(ClassError):
        def __init__(self, cls_instance, message):
            super().__init__(
                cls_instance=cls_instance,
                message=message,
                entity_id=cls_instance.auth.domo_instance,
            )

    async def get_is_invite_social_users_enabled(
        self,
        customer_id: str = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
    ):
        """checks if users can be invited as social users to the instaance"""

        if not customer_id:
            from .bootstrap import DomoBootstrap

            try:
                bs = DomoBootstrap(auth=self.auth)
                customer_id = await bs.get_customer_id()

            except InvalidAuthTypeError as e:
                raise self.InstanceConfig_ClassError(
                    self,
                    message=f"{e.__class__.__name__} -- bootstrap API requires FullAuth",
                ) from e

        res = await toggle_routes.get_is_invite_social_users_enabled(
            auth=self.auth,
            customer_id=customer_id,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        self.is_invite_social_users_enabled = res.response["is_enabled"]

        if return_raw:
            return res

        return res.response

    async def toggle_is_invite_social_users_enabled(
        self,
        is_enabled: bool,
        customer_id: str = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):
        """enables or disables the ability to invite users to instance as social users"""

        # Get customer_id if not provided
        if not customer_id:
            from .bootstrap import DomoBootstrap

            try:
                bs = DomoBootstrap(auth=self.auth)
                customer_id = await bs.get_customer_id()

            except InvalidAuthTypeError as e:
                raise self.InstanceConfig_ClassError(
                    self,
                    message=f"{e.__class__.__name__} -- bootstrap API requires FullAuth",
                ) from e

        res_is_enabled = await self.get_is_invite_social_users_enabled(
            customer_id=customer_id
        )

        if is_enabled == self.is_invite_social_users_enabled:
            return res_is_enabled

        res = await toggle_routes.toggle_is_invite_social_users_enabled(
            auth=self.auth,
            customer_id=customer_id,
            is_enabled=is_enabled,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        res_is_enabled = await self.get_is_invite_social_users_enabled()

        if return_raw:
            return res

        return res_is_enabled

    async def get_is_weekly_digest_enabled(
        self,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        """the weekly digest is a weekly email from Domo of changes to the instance"""

        res = await toggle_routes.get_is_weekly_digest_enabled(
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        self.is_weekly_digest_enabled = res.response["is_enabled"]

        return res.response

    async def toggle_is_weekly_digest_enabled(
        self,
        is_enabled: bool,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_prn: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        """toggles if weekly digest is enabled or disabled"""

        res_is_enabled = await self.get_is_weekly_digest_enabled()

        if is_enabled == self.is_weekly_digest_enabled:
            if debug_prn:
                print(
                    f"weekly digest is already {'enabled' if is_enabled else 'disabled'} in {self.auth.domo_instance}"
                )
            return res_is_enabled

        if debug_prn:
            print(
                f"{'enabling' if is_enabled else 'disabling'} weekly digest {self.auth.domo_instance}"
            )

        res = await toggle_routes.toggle_is_weekly_digest_enabled(
            auth=self.auth,
            is_enabled=is_enabled,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        res_is_enabled = await self.get_is_weekly_digest_enabled()

        if return_raw:
            return res

        return res_is_enabled

    async def toggle_is_left_nav_enabled(
        self,
        is_use_left_nav: bool = True,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=1,
    ):
        """toggles the use of the left nav in Domo"""

        res = await toggle_routes.toggle_is_left_nav_enabled(
            auth=self.auth,
            is_use_left_nav=is_use_left_nav,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.is_use_left_nav = res.response["is_enabled"]

        return res

    async def get_is_left_nav_enabled(
        self,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=1,
    ):
        """gets the use of the left nav in Domo"""

        res = await toggle_routes.get_is_left_nav_enabled(
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.is_use_left_nav = res.response["is_enabled"]

        return res



"""DomoEverywhere - Domo Publication and Subscription Management

This module provides classes for managing Domo Everywhere publications,
subscriptions, and content distribution across Domo instances.

Classes:
    DomoEverywhere: Main class for managing publications and subscriptions
    DomoPublication: Publication entity management
    DomoSubscription: Subscription entity management
    DomoPublication_Content: Content within publications
    DomoPublication_Content_Enum: Enum of publishable content types

Example:
    Basic DomoEverywhere usage:

        >>> from domolibrary2.classes.DomoEverywhere import DomoEverywhere
        >>> de = DomoEverywhere(auth=auth)
        >>> publications = await de.get_publications()
        >>> subscriptions = await de.get_subscriptions()
"""

from .core import (
    DomoEverywhere,
    DomoPublication,
    DomoPublication_Content,
    DomoPublication_Content_Enum,
    DomoPublication_UnexpectedContentType,
    DomoSubscription,
    DomoSubscription_NoParent,
    DomoSubscription_NoParentAuth,
)

__all__ = [
    "DomoEverywhere",
    "DomoPublication",
    "DomoPublication_Content",
    "DomoPublication_Content_Enum",
    "DomoPublication_UnexpectedContentType",
    "DomoSubscription",
    "DomoSubscription_NoParent",
    "DomoSubscription_NoParentAuth",
]



"""
Unified Access Control System for Domo Objects

This module provides a comprehensive access control framework that intelligently maps
users and groups to various Domo objects with different permission levels.

The system handles access control for:
- DomoAccounts (credential sharing)
- DomoCards (content sharing)
- DomoDatasets (data access)
- DomoGroups (membership management)
- PDP Policies (data row-level security)
- DomoPages (dashboard sharing)

Key Concepts:
1. **Access Levels**: Standardized permission levels (OWNER, ADMIN, EDITOR, VIEWER, etc.)
2. **Access Types**: Different types of access (DIRECT, INHERITED, POLICY_BASED)
3. **Entity Relationships**: Unified representation of user/group relationships to objects
4. **Access Inheritance**: How permissions cascade through groups and hierarchies
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any

from ...auth import DomoAuth
from ...base.base import DomoEnum
from ...base.entities import DomoEntity
from ...base.exceptions import DomoError


class AccessLevel(DomoEnum):
    """Standardized access levels across all Domo objects."""

    OWNER = "OWNER"  # Full control including deletion
    ADMIN = "ADMIN"  # Administrative access, can manage permissions
    EDITOR = "EDITOR"  # Can modify content but not permissions
    CONTRIBUTOR = "CONTRIBUTOR"  # Can add content but not modify existing
    VIEWER = "VIEWER"  # Read-only access
    NONE = "NONE"  # No access

    default = NONE


class EntityType(DomoEnum):
    """Types of entities that can have access."""

    USER = "USER"
    GROUP = "GROUP"
    ROLE = "ROLE"

    default = USER


@dataclass
class AccessGrant:
    """Represents a single access grant to an entity."""

    entity_id: str
    entity_type: EntityType
    access_level: AccessLevel
    granted_by: str | None = None  # ID of who granted access
    granted_date: str | None = None
    effective_date: str | None = None
    expiry_date: str | None = None

    # Optional: track if this came from group membership resolution
    source_group_id: str | None = None  # Group ID if resolved from group

    # Additional metadata
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class AccessSummary:
    """Summary of all access grants for a specific entity."""

    entity_id: str
    entity_type: EntityType
    effective_access_level: AccessLevel  # Highest level from all grants
    access_grants: list[AccessGrant] = field(default_factory=list)

    @property
    def direct_access_level(self) -> AccessLevel:
        """Get the highest direct access level (not from group resolution)."""
        direct_grants = [g for g in self.access_grants if g.source_group_id is None]
        if not direct_grants:
            return AccessLevel.NONE

        # Return highest access level (assuming enum order represents hierarchy)
        access_levels = [g.access_level for g in direct_grants]
        return max(access_levels, key=lambda x: list(AccessLevel).index(x))

    @property
    def group_access_level(self) -> AccessLevel:
        """Get the highest access level from group membership."""
        group_grants = [g for g in self.access_grants if g.source_group_id is not None]
        if not group_grants:
            return AccessLevel.NONE

        access_levels = [g.access_level for g in group_grants]
        return max(access_levels, key=lambda x: list(AccessLevel).index(x))


class DomoAccessController(ABC):
    """Abstract base class for object-specific access controllers."""

    def __init__(self, auth: DomoAuth, parent_object: DomoEntity):
        self.auth = auth
        self.parent_object = parent_object
        self.parent_id = parent_object.id

        # Cache for access data
        self._access_cache: dict[str, AccessSummary] = {}
        self._cache_valid = False

    @abstractmethod
    async def get_direct_access_grants(self) -> list[AccessGrant]:
        """Get all direct access grants for the parent object."""
        pass

    @abstractmethod
    async def grant_access(
        self,
        entity_id: str,
        entity_type: EntityType,
        access_level: AccessLevel,
        **kwargs,
    ) -> bool:
        """Grant access to an entity."""
        pass

    @abstractmethod
    async def revoke_access(
        self, entity_id: str, entity_type: EntityType, **kwargs
    ) -> bool:
        """Revoke access from an entity."""
        pass

    async def get_inherited_access_grants(self) -> list[AccessGrant]:
        """Get inherited access grants (through group memberships)."""
        # This would query group memberships and calculate inherited permissions
        # Implementation depends on group hierarchy and membership resolution
        inherited_grants = []

        # Get all direct grants for groups
        direct_grants = await self.get_direct_access_grants()
        group_grants = [g for g in direct_grants if g.entity_type == EntityType.GROUP]

        # For each group, get its members and create inherited grants
        for group_grant in group_grants:
            group_members = await self._get_group_members(group_grant.entity_id)

            for member in group_members:
                inherited_grant = AccessGrant(
                    entity_id=member.id,
                    entity_type=(
                        EntityType.USER
                        if hasattr(member, "email")
                        else EntityType.GROUP
                    ),
                    access_level=group_grant.access_level,
                    source_group_id=group_grant.entity_id,
                )
                inherited_grants.append(inherited_grant)

        return inherited_grants

    async def get_all_access_grants(
        self, include_inherited: bool = True
    ) -> list[AccessGrant]:
        """Get all access grants (direct and optionally inherited)."""
        grants = await self.get_direct_access_grants()

        if include_inherited:
            inherited_grants = await self.get_inherited_access_grants()
            grants.extend(inherited_grants)

        return grants

    async def get_access_summary(self, entity_id: str) -> AccessSummary:
        """Get comprehensive access summary for a specific entity."""
        if not self._cache_valid or entity_id not in self._access_cache:
            await self._refresh_access_cache()

        return self._access_cache.get(
            entity_id,
            AccessSummary(
                entity_id=entity_id,
                entity_type=EntityType.USER,  # Default, should be determined
                effective_access_level=AccessLevel.NONE,
            ),
        )

    async def get_all_access_summaries(self) -> dict[str, AccessSummary]:
        """Get access summaries for all entities with access."""
        if not self._cache_valid:
            await self._refresh_access_cache()

        return self._access_cache.copy()

    async def _refresh_access_cache(self):
        """Refresh the internal access cache."""
        all_grants = await self.get_all_access_grants()

        # Group grants by entity
        entity_grants: dict[str, list[AccessGrant]] = {}
        for grant in all_grants:
            if grant.entity_id not in entity_grants:
                entity_grants[grant.entity_id] = []
            entity_grants[grant.entity_id].append(grant)

        # Create access summaries
        self._access_cache = {}
        for entity_id, grants in entity_grants.items():
            # Calculate effective access level (highest level from all grants)
            access_levels = [g.access_level for g in grants]
            effective_level = max(
                access_levels, key=lambda x: list(AccessLevel).index(x)
            )

            # Determine entity type from grants
            entity_type = grants[0].entity_type if grants else EntityType.USER

            summary = AccessSummary(
                entity_id=entity_id,
                entity_type=entity_type,
                effective_access_level=effective_level,
                access_grants=grants,
            )
            self._access_cache[entity_id] = summary

        self._cache_valid = True

    async def _get_group_members(self, group_id: str) -> list[Any]:
        """Get members of a group. Override in subclasses for specific implementations."""
        # This would be implemented to get actual group members
        # For now, return empty list
        return []

    def invalidate_cache(self):
        """Invalidate the access cache to force refresh on next access."""
        self._cache_valid = False


class DomoObjectAccessManager:
    """Main access manager that provides unified access control across all Domo objects."""

    def __init__(self, auth: DomoAuth):
        self.auth = auth
        self._controllers: dict[str, DomoAccessController] = {}

    def register_controller(self, object_type: str, controller_class: type):
        """Register an access controller for a specific object type."""
        self._controllers[object_type] = controller_class

    def get_controller(self, domo_object: DomoEntity) -> DomoAccessController | None:
        """Get the appropriate access controller for a Domo object."""
        object_type = type(domo_object).__name__

        if object_type in self._controllers:
            return self._controllers[object_type](self.auth, domo_object)

        return None

    async def get_user_access_across_objects(
        self, user_id: str, object_types: list[str] | None = None
    ) -> dict[str, dict[str, AccessSummary]]:
        """Get a user's access across all objects of specified types."""
        # This would query all objects of specified types and check user access
        # Implementation would depend on available APIs and object discovery methods
        pass

    async def get_object_access_summary(
        self, domo_object: DomoEntity, include_inherited: bool = True
    ) -> dict[str, AccessSummary]:
        """Get complete access summary for a Domo object."""
        controller = self.get_controller(domo_object)
        if not controller:
            raise DomoError(
                f"No access controller found for object type: {type(domo_object).__name__}"
            )

        return await controller.get_all_access_summaries()


# Concrete implementations for specific object types


class DomoAccountAccessController(DomoAccessController):
    """Access controller for DomoAccount objects."""

    async def get_direct_access_grants(self) -> list[AccessGrant]:
        """Get direct access grants for the account."""
        from ...routes import account as account_routes

        res = await account_routes.get_account_accesslist(
            auth=self.auth, account_id=self.parent_id
        )

        grants = []
        for item in res.response:
            grant = AccessGrant(
                entity_id=item["id"],
                entity_type=(
                    EntityType.USER if item["type"] == "USER" else EntityType.GROUP
                ),
                access_level=self._map_account_access_level(item["accessLevel"]),
            )
            grants.append(grant)

        return grants

    async def grant_access(
        self,
        entity_id: str,
        entity_type: EntityType,
        access_level: AccessLevel,
        **kwargs,
    ) -> bool:
        """Grant account access to an entity."""
        # Implementation would use account sharing routes
        pass

    async def revoke_access(
        self, entity_id: str, entity_type: EntityType, **kwargs
    ) -> bool:
        """Revoke account access from an entity."""
        # Implementation would use account unsharing routes
        pass

    def _map_account_access_level(self, domo_access_level: str) -> AccessLevel:
        """Map Domo account access levels to standardized access levels."""
        mapping = {
            "OWNER": AccessLevel.OWNER,
            "ADMIN": AccessLevel.ADMIN,
            "EDIT": AccessLevel.EDITOR,
            "VIEW": AccessLevel.VIEWER,
        }
        return mapping.get(domo_access_level, AccessLevel.NONE)


class DomoGroupAccessController(DomoAccessController):
    """Access controller for DomoGroup objects (membership management)."""

    async def get_direct_access_grants(self) -> list[AccessGrant]:
        """Get direct membership grants for the group."""
        from ...routes import group as group_routes

        # Get owners
        owners_res = await group_routes.get_group_owners(
            auth=self.auth, group_id=self.parent_id
        )

        # Get members
        members_res = await group_routes.get_group_membership(
            auth=self.auth, group_id=self.parent_id
        )

        grants = []

        # Process owners
        for owner in owners_res.response:
            grant = AccessGrant(
                entity_id=owner.get("userId") or owner.get("id"),
                entity_type=(
                    EntityType.USER
                    if owner.get("type", "USER") == "USER"
                    else EntityType.GROUP
                ),
                access_level=AccessLevel.OWNER,
            )
            grants.append(grant)

        # Process members
        for member in members_res.response:
            grant = AccessGrant(
                entity_id=member.get("userId") or member.get("id"),
                entity_type=(
                    EntityType.USER
                    if member.get("type", "USER") == "USER"
                    else EntityType.GROUP
                ),
                access_level=AccessLevel.VIEWER,  # Regular members
            )
            grants.append(grant)

        return grants

    async def grant_access(
        self,
        entity_id: str,
        entity_type: EntityType,
        access_level: AccessLevel,
        **kwargs,
    ) -> bool:
        """Add entity to group with specified access level."""
        # Implementation would use group membership routes
        pass

    async def revoke_access(
        self, entity_id: str, entity_type: EntityType, **kwargs
    ) -> bool:
        """Remove entity from group."""
        # Implementation would use group membership routes
        pass


__all__ = [
    "AccessLevel",
    "EntityType",
    "AccessGrant",
    "AccessSummary",
    "DomoAccessController",
    "DomoObjectAccessManager",
    "DomoAccountAccessController",
    "DomoGroupAccessController",
]



"""Jupyter authentication classes for Domo."""

from dataclasses import dataclass, field

from .base import _DomoAuth_Optional, _DomoAuth_Required
from .full import DomoFullAuth, _DomoFullAuth_Required
from .token import DomoTokenAuth, _DomoTokenAuth_Required


class _DomoJupyter_Required:  # noqa: N801
    """Required parameters and setup for Domo Jupyter authentication.

    This class provides the foundational authentication components needed for
    Domo Jupyter environments, including token validation and environment setup.

    Attributes:
        jupyter_token (str): Authorization token from Domo Jupyter network traffic
        service_location (str): Service location from Domo Jupyter environment
        service_prefix (str): Service prefix from Domo Jupyter environment
    """

    def __init__(
        self,
        jupyter_token: str | None = None,
        service_location: str | None = None,
        service_prefix: str | None = None,
    ):
        """Initialize Jupyter authentication parameters.

        If parameters are not provided, the user will be prompted to enter them
        interactively. These values are typically obtained by monitoring Domo
        Jupyter network traffic.

        Args:
            jupyter_token (str | None): Authorization token from network traffic
            service_location (str | None): Service location from environment
            service_prefix (str | None): Service prefix from environment

        Raises:
            ValueError: If any required parameters are missing after initialization
        """

        self.jupyter_token = jupyter_token or input(
            "jupyter token: # retrieve this by monitoring Domo Jupyter network traffic. It is the Authorization header\n> "
        )
        self.service_location = service_location or input(
            "service_location: # retrieve from Domo Jupyter environment\n> "
        )
        self.service_prefix = service_prefix or input(
            "service_prefix: # retrieve from Domo Jupyter environment\n> "
        )

        self._test_prereq()

    def get_jupyter_token_flow(self):
        """Stub method for initiating a Jupyter token retrieval flow.

        This method is a placeholder for future implementation of automated
        Jupyter token retrieval.
        """
        print("hello world, I am a jupyter_token")

    def _test_prereq(self):
        """Validate that required attributes are present.

        Raises:
            ValueError: If any required Jupyter parameters are missing
        """
        missing = []
        if not self.jupyter_token:
            missing.append("jupyter_token")
        if not self.service_location:
            missing.append("service_location")
        if not self.service_prefix:
            missing.append("service_prefix")

        if missing:
            raise ValueError(f"DomoJupyterAuth objects must have: {', '.join(missing)}")


@dataclass
class DomoJupyterAuth(_DomoAuth_Optional, _DomoJupyter_Required, _DomoAuth_Required):
    """Base class for Domo Jupyter authentication.

    This class combines the core authentication functionality with Jupyter-specific
    requirements. It serves as a foundation for specific Jupyter authentication types.
    """


@dataclass
class DomoJupyterFullAuth(
    _DomoJupyter_Required,
    _DomoFullAuth_Required,
):
    """Jupyter authentication using full credentials (username/password).

    This class combines full Domo authentication with Jupyter environment support.
    It's used when working within Domo Jupyter environments and requires both
    standard Domo credentials and Jupyter-specific authentication tokens.

    Attributes:
        jupyter_token (str): Authorization token from Domo Jupyter (not shown in repr)
        service_location (str): Service location from Jupyter environment
        service_prefix (str): Service prefix from Jupyter environment
        domo_instance (str): The Domo instance identifier
        domo_username (str): Domo username for authentication
        domo_password (str): Domo password for authentication (not shown in repr)
        token_name (str | None): Name identifier for the token
        token (str | None): The authentication token (not shown in repr)
        user_id (str | None): The authenticated user's ID
        is_valid_token (bool): Whether the current token is valid

    Example:
        >>> auth = DomoJupyterFullAuth(
        ...     jupyter_token="jupyter-auth-token",
        ...     service_location="service-location",
        ...     service_prefix="service-prefix",
        ...     domo_instance="mycompany",
        ...     domo_username="user@company.com",
        ...     domo_password="secure_password"
        ... )
    """

    jupyter_token: str = field(repr=False)
    service_location: str
    service_prefix: str

    domo_instance: str
    domo_username: str
    domo_password: str = field(repr=False)

    token_name: str | None = None
    token: str | None = field(default=None, repr=False)
    user_id: str | None = None
    is_valid_token: bool = False

    def __post_init__(self):
        """Initialize the Jupyter and full authentication mixins."""
        _DomoJupyter_Required.__init__(
            self,
            jupyter_token=self.jupyter_token,
            service_location=self.service_location,
            service_prefix=self.service_prefix,
        )
        _DomoFullAuth_Required.__init__(
            self,
            domo_username=self.domo_username,
            domo_password=self.domo_password,
            domo_instance=self.domo_instance,
            token_name=self.token_name,
            token=self.token,
            user_id=self.user_id,
            is_valid_token=self.is_valid_token,
        )

    @property
    def auth_header(self) -> dict:
        """Generate authentication headers combining Domo and Jupyter tokens.

        Returns:
            dict: Combined authentication headers for both Domo and Jupyter APIs
        """
        return {
            **super().auth_header,
            "authorization": f"Token {self.jupyter_token}",
        }

    @classmethod
    def convert_auth(
        cls, auth: DomoFullAuth, jupyter_token, service_location, service_prefix
    ):
        """Convert DomoFullAuth to DomoJupyterFullAuth.

        This factory method creates a Jupyter-enabled authentication object from
        an existing full authentication object by adding the necessary Jupyter
        environment parameters.

        Args:
            auth (DomoFullAuth): Existing full authentication object
            jupyter_token (str): Authorization token from Domo Jupyter
            service_location (str): Service location from Jupyter environment
            service_prefix (str): Service prefix from Jupyter environment

        Returns:
            DomoJupyterFullAuth: New Jupyter-enabled authentication object
        """
        return cls(
            domo_instance=auth.domo_instance,
            domo_username=auth.domo_username,
            domo_password=auth.domo_password,
            jupyter_token=jupyter_token,
            service_location=service_location,
            service_prefix=service_prefix,
            token_name=auth.token_name,
            token=auth.token,
            user_id=auth.user_id,
            is_valid_token=auth.is_valid_token,
        )


@dataclass
class DomoJupyterTokenAuth(
    _DomoJupyter_Required,
    _DomoTokenAuth_Required,
):
    """Jupyter authentication using access tokens.

    This class combines token-based Domo authentication with Jupyter environment support.
    It's used when working within Domo Jupyter environments with pre-generated access
    tokens instead of username/password credentials.

    Attributes:
        jupyter_token (str): Authorization token from Domo Jupyter (not shown in repr)
        service_location (str): Service location from Jupyter environment
        service_prefix (str): Service prefix from Jupyter environment
        domo_instance (str): The Domo instance identifier
        domo_access_token (str): Pre-generated access token (not shown in repr)
        token_name (str | None): Name identifier for the token
        token (str | None): The authentication token (not shown in repr)
        user_id (str | None): The authenticated user's ID
        is_valid_token (bool): Whether the current token is valid

    Example:
        >>> auth = DomoJupyterTokenAuth(
        ...     jupyter_token="jupyter-auth-token",
        ...     service_location="service-location",
        ...     service_prefix="service-prefix",
        ...     domo_instance="mycompany",
        ...     domo_access_token="your-access-token"
        ... )
    """

    jupyter_token: str = field(repr=False)
    service_location: str
    service_prefix: str

    domo_instance: str

    domo_access_token: str = field(repr=False)

    token_name: str | None = None
    token: str | None = field(default=None, repr=False)
    user_id: str | None = None
    is_valid_token: bool = False

    def __post_init__(self):
        """Initialize the Jupyter and token authentication mixins."""
        _DomoJupyter_Required.__init__(
            self,
            jupyter_token=self.jupyter_token,
            service_location=self.service_location,
            service_prefix=self.service_prefix,
        )
        _DomoTokenAuth_Required.__init__(
            self,
            domo_access_token=self.domo_access_token,
            domo_instance=self.domo_instance,
            token_name=self.token_name,
            token=self.token,
            user_id=self.user_id,
            is_valid_token=self.is_valid_token,
        )

    @property
    def auth_header(self) -> dict:
        """Generate authentication headers combining Domo and Jupyter tokens.

        Returns:
            dict: Combined authentication headers for both Domo and Jupyter APIs
        """
        return {
            **super().auth_header,
            "authorization": f"Token {self.jupyter_token}",
        }

    @classmethod
    def convert_auth(
        cls, auth: DomoTokenAuth, jupyter_token, service_location, service_prefix
    ):
        """Convert DomoTokenAuth to DomoJupyterTokenAuth.

        This factory method creates a Jupyter-enabled authentication object from
        an existing token authentication object by adding the necessary Jupyter
        environment parameters.

        Args:
            auth (DomoTokenAuth): Existing token authentication object
            jupyter_token (str): Authorization token from Domo Jupyter
            service_location (str): Service location from Jupyter environment
            service_prefix (str): Service prefix from Jupyter environment

        Returns:
            DomoJupyterTokenAuth: New Jupyter-enabled authentication object
        """
        return cls(
            domo_instance=auth.domo_instance,
            domo_access_token=auth.domo_access_token,
            jupyter_token=jupyter_token,
            service_location=service_location,
            service_prefix=service_prefix,
            token_name=auth.token_name,
            token=auth.token,
            user_id=auth.user_id,
            is_valid_token=auth.is_valid_token,
        )



"""A class-based approach for interpreting Domo schedule configurations"""

__all__ = [
    "DomoSchedule_Base",
    "DomoCronSchedule",
    "DomoManualSchedule",
    "ScheduleFrequencyEnum",
    "ScheduleType",
]

import datetime as dt
import json
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable

# from ...auth import DomoAuth
from ...base.base import DomoBase, DomoEnumMixin


class ScheduleFrequencyEnum(DomoEnumMixin, Enum):
    """Common schedule frequency types"""

    MANUAL = "MANUAL"
    ONCE = "ONCE"
    MINUTELY = "MINUTELY"
    HOURLY = "HOURLY"
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"
    YEARLY = "YEARLY"
    CUSTOM_CRON = "CUSTOM_CRON"


class ScheduleType(DomoEnumMixin, Enum):
    """Schedule configuration types"""

    MANUAL = "MANUAL"
    CRON = "CRON"


@dataclass
class DomoSchedule_Base(DomoBase, ABC):
    """Base class for interpreting and managing Domo schedule configurations"""

    # Raw schedule data
    schedule_start_date: dt.datetime | None = None

    # Interpreted schedule information
    frequency: ScheduleFrequencyEnum = ScheduleFrequencyEnum.MANUAL
    schedule_type: ScheduleType = ScheduleType.MANUAL

    # Detailed frequency information
    interval: int = 1
    minute: int | None = None
    hour: int | None = None
    day_of_week: list[int] | None = None  # 0=Sunday, 6=Saturday
    day_of_month: list[int] | None = None
    month: list[int] | None = None

    # Schedule metadata
    timezone: str | None = None
    is_active: bool = True
    next_run_time: dt.datetime | None = None

    # Raw data for reference
    raw: dict[str, Any] = field(default_factory=dict, repr=False)

    parent: Any = field(
        default=None, repr=None
    )  # DomoDataset, DomoStream, DomoDataflow

    def __post_init__(self):
        """Post-initialization to interpret the schedule configuration"""
        self._interpret_schedule()

    @abstractmethod
    def _interpret_schedule(self):
        """Interpret the schedule configuration from raw data - implemented by subclasses"""
        pass

    @staticmethod
    def _extract_schedule_components(schedule_data: dict[str, Any]) -> dict[str, Any]:
        """Extract schedule components from advanced JSON configuration"""
        components = {
            "timezone": schedule_data.get("timezone"),
            "interval": schedule_data.get("interval", 1),
            "minute": schedule_data.get("minute"),
            "hour": schedule_data.get("hour"),
            "day_of_week": schedule_data.get("daysOfWeek"),
            "day_of_month": schedule_data.get("daysOfMonth"),
            "month": schedule_data.get("months"),
        }

        # Extended parsing for 'type' and 'at' fields
        type_map = {
            "MINUTE": ScheduleFrequencyEnum.MINUTELY,
            "HOUR": ScheduleFrequencyEnum.HOURLY,
            "DAY": ScheduleFrequencyEnum.DAILY,
            "WEEK": ScheduleFrequencyEnum.WEEKLY,
            "MONTH": ScheduleFrequencyEnum.MONTHLY,
            "YEAR": ScheduleFrequencyEnum.YEARLY,
        }
        if "type" in schedule_data:
            type_str = str(schedule_data["type"]).upper()
            components["frequency"] = type_map.get(
                type_str, ScheduleFrequencyEnum.CUSTOM_CRON
            )
        else:
            components["frequency"] = ScheduleFrequencyEnum.CUSTOM_CRON

        # Parse 'at' field for hour and minute
        at_str = schedule_data.get("at")
        if at_str:
            # Expect format like '09:02 AM' or '14:30'
            import re

            match = re.match(
                r"(\d{1,2}):(\d{2})\s*(AM|PM)?", at_str.strip(), re.IGNORECASE
            )
            if match:
                hour = int(match.group(1))
                minute = int(match.group(2))
                ampm = match.group(3)
                if ampm:
                    if ampm.upper() == "PM" and hour < 12:
                        hour += 12
                    elif ampm.upper() == "AM" and hour == 12:
                        hour = 0
                components["hour"] = hour
                components["minute"] = minute

        # Determine frequency from explicit 'frequency' field if present
        if schedule_data.get("frequency"):
            freq_str = schedule_data["frequency"].upper()
            if freq_str in [f.value for f in ScheduleFrequencyEnum]:
                components["frequency"] = ScheduleFrequencyEnum(freq_str)

        return components

    @staticmethod
    def _normalize_expression(expression: str) -> str:
        """Normalize schedule expression for parsing"""
        return expression.upper().strip() if expression else ""

    @staticmethod
    def _detect_expression_type(
        expr: str,
    ) -> tuple[ScheduleFrequencyEnum, ScheduleType]:
        """Detect the type and frequency from expression"""
        # Manual execution
        if expr in ["MANUAL", "NONE", ""]:
            return ScheduleFrequencyEnum.MANUAL, ScheduleType.MANUAL

        # Once execution
        if expr in ["ONCE", "RUN_ONCE"]:
            return ScheduleFrequencyEnum.ONCE, ScheduleType.CRON

        # Default for expressions that need further parsing
        return ScheduleFrequencyEnum.CUSTOM_CRON, ScheduleType.CRON

    def _is_cron_expression(self, expr: str) -> bool:
        """Check if expression looks like a cron expression"""
        # Basic cron has 5-6 fields separated by spaces
        parts = expr.split()
        return len(parts) >= 5 and len(parts) <= 6

    @staticmethod
    def _parse_cron_components(expr: str) -> dict[str, Any]:
        """Parse cron expression components and return parsed data"""
        parts = expr.split()

        if len(parts) < 5:
            return {"frequency": ScheduleFrequencyEnum.CUSTOM_CRON}

        minute_part, hour_part, day_month_part, month_part, day_week_part = parts[:5]

        result = {
            "frequency": ScheduleFrequencyEnum.CUSTOM_CRON,
            "minute": None,
            "hour": None,
        }

        # Extract specific numeric values
        try:
            if minute_part.isdigit():
                result["minute"] = int(minute_part)
            if hour_part.isdigit():
                result["hour"] = int(hour_part)
        except ValueError:
            pass

        # Infer frequency from pattern
        if minute_part != "*" and hour_part == "*":
            result["frequency"] = ScheduleFrequencyEnum.HOURLY
        elif minute_part != "*" and hour_part != "*" and day_month_part == "*":
            result["frequency"] = ScheduleFrequencyEnum.DAILY
        elif day_week_part != "*":
            result["frequency"] = ScheduleFrequencyEnum.WEEKLY
        elif day_month_part != "*":
            result["frequency"] = ScheduleFrequencyEnum.MONTHLY

        return result

    def _parse_cron_expression(self, expr: str):
        """Parse a cron expression (simplified)"""
        cron_data = self._parse_cron_components(expr)

        # Apply parsed data to instance
        self.frequency = cron_data["frequency"]
        if cron_data["minute"] is not None:
            self.minute = cron_data["minute"]
        if cron_data["hour"] is not None:
            self.hour = cron_data["hour"]

    @staticmethod
    def _parse_simple_expression_components(expr: str) -> dict[str, Any]:
        """Parse simple schedule expression and return components"""
        expr_lower = expr.lower()

        # Define patterns for different time units
        patterns = [
            (r"(\d+)\s*minute", ScheduleFrequencyEnum.MINUTELY),
            (r"(\d+)\s*hour", ScheduleFrequencyEnum.HOURLY),
            (r"(\d+)\s*day", ScheduleFrequencyEnum.DAILY),
            (r"(\d+)\s*week", ScheduleFrequencyEnum.WEEKLY),
            (r"(\d+)\s*month", ScheduleFrequencyEnum.MONTHLY),
        ]

        # Check for daily patterns
        if "daily" in expr_lower:
            match = re.search(r"(\d+)\s*daily", expr_lower)
            interval = int(match.group(1)) if match else 1
            return {"frequency": ScheduleFrequencyEnum.DAILY, "interval": interval}

        # Check other patterns
        for pattern, frequency in patterns:
            if frequency.value.lower()[:-2] in expr_lower:  # Remove 'LY' suffix
                match = re.search(pattern, expr_lower)
                interval = int(match.group(1)) if match else 1
                return {"frequency": frequency, "interval": interval}

        return {"frequency": ScheduleFrequencyEnum.CUSTOM_CRON, "interval": 1}

    def _parse_simple_expression(self, expr: str):
        """Parse simple schedule expressions"""
        components = self._parse_simple_expression_components(expr)

        self.frequency = components["frequency"]
        self.interval = components["interval"]

    def _infer_frequency_from_components(self):
        """Infer frequency based on available time components"""
        if self.day_of_week is not None:
            self.frequency = ScheduleFrequencyEnum.WEEKLY
        elif self.day_of_month is not None:
            self.frequency = ScheduleFrequencyEnum.MONTHLY
        elif self.month is not None:
            self.frequency = ScheduleFrequencyEnum.YEARLY
        elif self.hour is not None:
            self.frequency = ScheduleFrequencyEnum.DAILY
        elif self.minute is not None:
            self.frequency = ScheduleFrequencyEnum.HOURLY
        else:
            self.frequency = ScheduleFrequencyEnum.CUSTOM_CRON

    @staticmethod
    def _extract_field_mappings(obj: dict[str, Any]) -> dict[str, Any]:
        """Extract and map field names from input dictionary"""
        return {
            "start_date_raw": obj.get("scheduleStartDate") or obj.get("startDate"),
            "schedule_expr": obj.get("scheduleExpression") or obj.get("expression"),
            "advanced_json": obj.get("advancedScheduleJson")
            or obj.get("advancedSchedule"),
            "timezone": obj.get("timezone"),
            "is_active": obj.get("isActive", True),
        }

    @staticmethod
    def _extract_scheduler_fields(obj: dict[str, Any]) -> dict[str, Any]:
        """Extract all scheduler-relevant fields for raw storage"""
        scheduler_fields = {}

        # Common scheduler field names to capture
        scheduler_field_names = [
            "scheduleExpression",
            "scheduleStartDate",
            "advancedScheduleJson",
            "scheduleRetryExpression",
            "scheduleRetryCount",
            "timezone",
            "isActive",
            "nextRunTime",
            "lastRunTime",
            "scheduleActive",
            "expression",
            "startDate",
            "advancedSchedule",
            "retryExpression",
            "retryCount",
        ]

        # Extract any scheduler-related fields that exist in obj
        for field_name in scheduler_field_names:
            if field_name in obj:
                scheduler_fields[field_name] = obj[field_name]

        return scheduler_fields

    @staticmethod
    def _parse_datetime_input(date_input: Any) -> dt.datetime | None:
        """Parse various datetime input formats into datetime object"""
        if not date_input:
            return None

        # String formats
        if isinstance(date_input, str):
            # Try ISO format first
            try:
                return dt.datetime.fromisoformat(date_input.replace("Z", "+00:00"))
            except ValueError:
                pass

            # Try standard format
            try:
                return dt.datetime.strptime(date_input, "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                pass

            # Try other common formats
            formats_to_try = [
                "%Y-%m-%d %H:%M:%S",
                "%Y-%m-%d",
                "%m/%d/%Y %H:%M:%S",
                "%m/%d/%Y",
            ]

            for fmt in formats_to_try:
                try:
                    return dt.datetime.strptime(date_input, fmt)
                except ValueError:
                    continue

        # Numeric timestamp (assuming milliseconds)
        elif isinstance(date_input, (int, float)):
            try:
                # Handle both seconds and milliseconds timestamps
                if date_input > 1e10:  # Likely milliseconds
                    return dt.datetime.fromtimestamp(date_input / 1000)
                else:  # Likely seconds
                    return dt.datetime.fromtimestamp(date_input)
            except (ValueError, OSError):
                pass

        return None

    @staticmethod
    def _parse_json_input(json_input: Any) -> dict[str, Any] | None:
        """Parse JSON input that might be string or dict"""
        if not json_input:
            return None

        if isinstance(json_input, dict):
            return json_input

        if isinstance(json_input, str):
            try:
                return json.loads(json_input)
            except json.JSONDecodeError:
                return None

        return None

    def to_dict(self, override_fn: Callable | None = None) -> dict[str, Any]:
        if override_fn:
            return override_fn(self)

        """Convert schedule to dictionary format"""
        result = super().to_dict()
        result.update(
            {
                "frequency": self.frequency.value,
                "scheduleType": self.schedule_type.value,
                "interval": self.interval,
                "isActive": self.is_active,
            }
        )

        if self.schedule_start_date:
            result["scheduleStartDate"] = self.schedule_start_date.isoformat()

        if self.day_of_week is not None:
            result["dayOfWeek"] = self.day_of_week

        if self.day_of_month is not None:
            result["dayOfMonth"] = self.day_of_month

        if self.next_run_time:
            result["nextRunTime"] = self.next_run_time.isoformat()

        return result

    def get_human_readable_schedule(self) -> str:
        """Get a human-readable description of the schedule"""
        if self.frequency == ScheduleFrequencyEnum.MANUAL:
            return "Manual execution"

        if self.frequency == ScheduleFrequencyEnum.ONCE:
            if self.schedule_start_date:
                return (
                    f"Run once on {self.schedule_start_date.strftime('%Y-%m-%d %H:%M')}"
                )
            return "Run once"

        base_desc = ""

        if self.frequency == ScheduleFrequencyEnum.MINUTELY:
            base_desc = f"Every {self.interval} minute(s)"
        elif self.frequency == ScheduleFrequencyEnum.HOURLY:
            base_desc = f"Every {self.interval} hour(s)"
            if self.minute is not None:
                base_desc += f" at {self.minute:02d} minutes past the hour"
        elif self.frequency == ScheduleFrequencyEnum.DAILY:
            base_desc = f"Every {self.interval} day(s)"
            if self.hour is not None and self.minute is not None:
                base_desc += f" at {self.hour:02d}:{self.minute:02d}"
        elif self.frequency == ScheduleFrequencyEnum.WEEKLY:
            base_desc = f"Every {self.interval} week(s)"
            if self.day_of_week:
                days = [
                    "Sunday",
                    "Monday",
                    "Tuesday",
                    "Wednesday",
                    "Thursday",
                    "Friday",
                    "Saturday",
                ]
                day_names = [days[day] for day in self.day_of_week if 0 <= day <= 6]
                base_desc += f" on {', '.join(day_names)}"
            if self.hour is not None and self.minute is not None:
                base_desc += f" at {self.hour:02d}:{self.minute:02d}"
        elif self.frequency == ScheduleFrequencyEnum.MONTHLY:
            base_desc = f"Every {self.interval} month(s)"
            if self.day_of_month:
                base_desc += f" on day(s) {', '.join(map(str, self.day_of_month))}"
            if self.hour is not None and self.minute is not None:
                base_desc += f" at {self.hour:02d}:{self.minute:02d}"
        elif self.frequency == ScheduleFrequencyEnum.YEARLY:
            base_desc = f"Every {self.interval} year(s)"
            if self.month:
                month_names = [
                    "Jan",
                    "Feb",
                    "Mar",
                    "Apr",
                    "May",
                    "Jun",
                    "Jul",
                    "Aug",
                    "Sep",
                    "Oct",
                    "Nov",
                    "Dec",
                ]
                month_names_list = [
                    month_names[m - 1] for m in self.month if 1 <= m <= 12
                ]
                base_desc += f" in {', '.join(month_names_list)}"
            if self.day_of_month:
                base_desc += f" on day(s) {', '.join(map(str, self.day_of_month))}"
        elif self.frequency == ScheduleFrequencyEnum.CUSTOM_CRON:
            base_desc = "Custom schedule"

        if self.timezone:
            base_desc += f" ({self.timezone})"

        if not self.is_active:
            base_desc += " [INACTIVE]"

        return base_desc

    def is_due_now(self, current_time: dt.datetime | None = None) -> bool:
        """Check if the schedule is due to run now (simplified logic)"""
        if not self.is_active or self.frequency == ScheduleFrequencyEnum.MANUAL:
            return False

        if current_time is None:
            current_time = dt.datetime.now()

        if self.schedule_start_date and current_time < self.schedule_start_date:
            return False

        # This is a simplified check - real implementation would need more complex logic
        # based on the specific schedule configuration
        return False  # Placeholder

    def __str__(self) -> str:
        return self.get_human_readable_schedule()

    def __repr__(self) -> str:
        return f"DomoSchedule(frequency={self.frequency.value}, type={self.schedule_type.value})"

    def export_as_dict(
        self, override_fn: Callable | None = None, return_snake_case: bool = False
    ) -> dict[str, Any]:
        """
        Export the schedule as a unified dictionary format.
        Args:
            override_fn: Optional function to override export logic.
            return_snake_case: If True, return keys in snake_case format.
        Returns:
            dict: Unified schedule dictionary.
        """
        if override_fn:
            return override_fn(self)

        # Unified schedule dict
        result = {
            "humanReadable": self.get_human_readable_schedule(),
            "frequency": self.frequency.value if self.frequency else None,
            "scheduleType": (self.schedule_type.value if self.schedule_type else None),
            "interval": self.interval,
            "minute": self.minute,
            "hour": self.hour,
            "dayOfWeek": self.day_of_week,
            "dayOfMonth": self.day_of_month,
            "month": self.month,
            "scheduleStartDate": (
                self.schedule_start_date.isoformat()
                if self.schedule_start_date
                else None
            ),
            "timezone": self.timezone,
            "isActive": self.is_active,
            "nextRunTime": (
                self.next_run_time.isoformat() if self.next_run_time else None
            ),
        }

        # Optionally include raw expression or advanced config if present
        if hasattr(self, "schedule_expression") and self.schedule_expression:
            result["expression"] = self.schedule_expression
        if hasattr(self, "advanced_schedule_json") and self.advanced_schedule_json:
            result["advancedScheduleJson"] = self.advanced_schedule_json

        # Convert to snake_case if requested
        if return_snake_case:
            from ...utils.convert import convert_str_to_snake_case

            result = {
                convert_str_to_snake_case(k, is_pascal=True): v
                for k, v in result.items()
            }

        return result

    async def refresh(
        self,
        session: Any | None = None,
        debug_api: bool = False,
        is_suppress_no_config: bool = False,
        **kwargs,
    ) -> "DomoSchedule_Base":
        """Refresh schedule data by refreshing the parent entity.

        This method calls the parent entity's refresh() method, which updates
        the parent's raw data. The schedule then re-initializes itself from
        the updated parent data.

        Note: If the schedule type changes (e.g., Manual to Cron), this method
        updates the instance in place. The parent entity should recreate the
        Schedule subentity if type safety is critical.

        Args:
            session: Optional httpx.AsyncClient session for API requests
            debug_api: Enable debug output for API calls
            is_suppress_no_config: Suppress errors for missing configs

        Returns:
            Self (with updated schedule data)

        Raises:
            AttributeError: If parent is None or doesn't support refresh
        """
        if not self.parent:
            raise AttributeError("Cannot refresh schedule: parent entity is not set")

        # Refresh the parent entity
        if hasattr(self.parent, "refresh") and callable(self.parent.refresh):
            await self.parent.refresh(
                session=session,
                debug_api=debug_api,
                is_suppress_no_config=is_suppress_no_config,
                **kwargs,
            )
        else:
            raise AttributeError(
                f"Parent entity {type(self.parent).__name__} does not have a refresh() method"
            )

        # Extract updated schedule data from parent's raw response
        parent_raw = getattr(self.parent, "raw", {})

        # Get scheduler fields from parent
        scheduler_raw = self._extract_scheduler_fields(parent_raw)
        self.raw = scheduler_raw

        # Re-extract and parse the updated schedule components
        field_mappings = self._extract_field_mappings(parent_raw)

        # Update schedule start date
        self.schedule_start_date = self._parse_datetime_input(
            field_mappings["start_date_raw"]
        )

        # Update basic fields
        self.timezone = field_mappings["timezone"]
        self.is_active = field_mappings["is_active"]

        # Update schedule expression if present
        if hasattr(self, "schedule_expression"):
            self.schedule_expression = field_mappings["schedule_expr"]  # type: ignore

        # Update advanced JSON if present - this may cause type to change
        advanced_json_raw = field_mappings["advanced_json"]
        if hasattr(self, "advanced_schedule_json"):
            self.advanced_schedule_json = self._parse_json_input(  # type: ignore
                advanced_json_raw
            )

        # Re-interpret the schedule to update frequency, interval, etc.
        # Note: For DomoManualSchedule, this will still set to MANUAL
        # For DomoCronSchedule, this will re-parse the cron/advanced config
        self._interpret_schedule()

        return self


@dataclass
class DomoManualSchedule(DomoSchedule_Base):
    """Schedule for manual execution only"""

    schedule_expression: str | None = None

    def _interpret_schedule(self):
        """Interpret manual schedule configuration"""
        self.schedule_type = ScheduleType.MANUAL
        self.frequency = ScheduleFrequencyEnum.MANUAL

    @classmethod
    def from_dict(
        cls, obj: dict[str, Any], parent: Any = None, **kwargs
    ) -> "DomoManualSchedule":
        """Create DomoManualSchedule from dictionary/API response"""
        field_mappings = cls._extract_field_mappings(obj)
        start_date = cls._parse_datetime_input(field_mappings["start_date_raw"])

        # Extract only scheduler-relevant fields for raw storage
        scheduler_raw = cls._extract_scheduler_fields(obj)

        return cls(
            parent=parent,
            schedule_start_date=start_date,
            schedule_expression=field_mappings["schedule_expr"],
            timezone=field_mappings["timezone"],
            is_active=field_mappings["is_active"],
            raw=scheduler_raw,
            **kwargs,
        )


@dataclass
class DomoCronSchedule(DomoSchedule_Base):
    """Schedule based on cron-like expressions, advanced JSON, or simple expressions"""

    schedule_expression: str | None = None
    advanced_schedule_json: dict[str, Any] | None = None

    def _interpret_schedule(self):
        """Interpret schedule from expression or advanced JSON"""
        self.schedule_type = ScheduleType.CRON

        # Priority 1: Advanced schedule JSON
        if self.advanced_schedule_json:
            self._interpret_advanced_schedule()
            return

        # Priority 2: Schedule expression
        if self.schedule_expression:
            self._interpret_expression_schedule()
            return

        # Default to manual if no configuration
        self.frequency = ScheduleFrequencyEnum.MANUAL

    def _interpret_advanced_schedule(self):
        """Interpret advanced schedule JSON configuration"""
        # Extract components using utility function
        components = self._extract_schedule_components(self.advanced_schedule_json)

        # Apply components to instance
        self.timezone = components["timezone"]
        self.interval = components["interval"]
        self.minute = components["minute"]
        self.hour = components["hour"]
        self.day_of_week = components["day_of_week"]
        self.day_of_month = components["day_of_month"]
        self.month = components["month"]
        self.frequency = components["frequency"]

        # Determine frequency based on available components if not explicitly set
        if components[
            "frequency"
        ] == ScheduleFrequencyEnum.CUSTOM_CRON and not self.advanced_schedule_json.get(
            "frequency"
        ):
            self._infer_frequency_from_components()

    def _interpret_expression_schedule(self):
        """Interpret cron-like or simple schedule expression"""
        expr = self._normalize_expression(self.schedule_expression)

        # Detect basic expression type
        frequency, _ = self._detect_expression_type(expr)
        self.frequency = frequency

        # If it's manual, return early
        if frequency == ScheduleFrequencyEnum.MANUAL:
            return

        # Try to parse cron expression (basic patterns)
        if self._is_cron_expression(expr):
            self._parse_cron_expression(expr)
        else:
            # Try simple frequency patterns
            self._parse_simple_expression(expr)

    @classmethod
    def from_dict(
        cls,
        obj: dict[str, Any],
        parent: Any | None = None,
        **kwargs,
    ) -> "DomoCronSchedule":
        """Create DomoCronSchedule from dictionary/API response"""
        field_mappings = cls._extract_field_mappings(obj)
        start_date = cls._parse_datetime_input(field_mappings["start_date_raw"])
        advanced_json = cls._parse_json_input(field_mappings["advanced_json"])

        # Extract only scheduler-relevant fields for raw storage
        scheduler_raw = cls._extract_scheduler_fields(obj)

        return cls(
            parent=parent,
            schedule_start_date=start_date,
            schedule_expression=field_mappings["schedule_expr"],
            advanced_schedule_json=advanced_json,
            timezone=field_mappings["timezone"],
            is_active=field_mappings["is_active"],
            raw=scheduler_raw,
            **kwargs,
        )


class DomoSchedule(DomoSchedule_Base):
    """Alias for DomoSchedule_Base to maintain backward compatibility"""

    @classmethod
    def determine_schedule_type(cls, obj: dict[str, Any]) -> type["DomoSchedule_Base"]:
        """Determine the appropriate schedule subclass based on input data"""
        field_mappings = cls._extract_field_mappings(obj)

        # Check for advanced schedule JSON and its type field
        advanced_json = cls._parse_json_input(field_mappings["advanced_json"])
        if advanced_json:
            schedule_type = advanced_json.get("type", "").upper()
            if schedule_type == "MANUAL":
                return DomoManualSchedule
            # Any other type in advanced JSON is handled by DomoCronSchedule
            return DomoCronSchedule

        # Check for schedule expression (cron-like)
        if field_mappings["schedule_expr"]:
            expr = cls._normalize_expression(field_mappings["schedule_expr"])
            frequency, schedule_type = cls._detect_expression_type(expr)

            if schedule_type == ScheduleType.MANUAL:
                return DomoManualSchedule
            else:
                return DomoCronSchedule

        # Default to manual schedule
        return DomoManualSchedule

    @classmethod
    def from_parent(
        cls, parent: Any, obj: dict | None = None, **kwargs
    ) -> "DomoSchedule_Base":
        """Create schedule instance from parent entity"""

        # If called on the base class, determine the correct subclass
        schedule_class = cls.determine_schedule_type(obj)
        return schedule_class.from_dict(obj, parent=parent, **kwargs)



__all__ = [
    "Group_GET_Error",
    "SearchGroups_Error",
    "search_groups_by_name",
    "get_all_groups",
    "get_group_by_id",
    "Group_CRUD_Error",
    "is_system_groups_visible",
    "toggle_system_group_visibility",
    "GroupType_Enum",
    "generate_body_create_group",
    "create_group",
    "update_group",
    "delete_groups",
    "get_group_owners",
    "get_group_membership",
    "generate_body_update_group_membership",
    "update_group_membership",
    "update_group_members",
    "update_group_owners",
]

from enum import Enum
from typing import Union

from ..auth import DomoAuth
from ..base import exceptions as dmde
from ..base.base import DomoEnumMixin
from ..client import (
    get_data as gd,
    response as rgd,
)
from ..client.context import RouteContext
from ..utils import convert as dmcv


class Group_GET_Error(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        message=None,
    ):
        super().__init__(res=res, message=message or res.response)


class SearchGroups_Error(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        message=None,
    ):
        super().__init__(res=res, message=message or res.response)


@gd.route_function
async def search_groups_by_name(
    auth: DomoAuth,
    search_name: str,
    is_exact_match: bool = True,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """uses /content/v2/groups/grouplist api -- includes user details"""
    context = RouteContext.build_context(context, **context_kwargs)

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/grouplist?ascending=true&search={search_name}&sort=name "

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )
    if not res.is_success:
        raise Group_GET_Error(res=res)

    if not is_exact_match:
        return res

    match_group = next(
        (group for group in res.response if group.get("name") == search_name), None
    )

    if not match_group:
        raise SearchGroups_Error(
            res=res,
            message=f"There is no exact match for {search_name}",
        )

    res.response = match_group

    return res


@gd.route_function
async def get_all_groups(
    auth: DomoAuth,
    *,
    context: RouteContext | None = None,
    debug_loop: bool = False,
    return_raw: bool = False,
    maximum=None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """uses /content/v2/groups/grouplist api -- includes user details"""
    context = RouteContext.build_context(context, **context_kwargs)

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/grouplist"

    def arr_fn(res):
        return res.response

    res = await gd.looper(
        offset_params={"offset": "offset", "limit": "limit"},
        arr_fn=arr_fn,
        loop_until_end=True if maximum is None else False,
        limit=30,
        url=url,
        method="GET",
        auth=auth,
        context=context,
        debug_loop=debug_loop,
        return_raw=return_raw,
        maximum=maximum,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Group_GET_Error(res=res)

    return res


@gd.route_function
async def get_group_by_id(
    auth: DomoAuth,
    group_id: str,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """uses /content/v2/groups/ api -- does not return details"""
    context = RouteContext.build_context(context, **context_kwargs)

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/{group_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        context=context,
    )

    if res.status == 404 and res.response == "Not Found":
        raise SearchGroups_Error(
            res=res,
            message=f"group {group_id} not found",
        )

    if not res.is_success:
        raise Group_GET_Error(res=res)

    return res


class Group_CRUD_Error(dmde.RouteError):
    def __init__(self, res: rgd.ResponseGetData, message=None):
        super().__init__(res=res, message=message or res.response)


@gd.route_function
async def is_system_groups_visible(
    auth: DomoAuth,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
):
    context = RouteContext.build_context(context, **context_kwargs)

    url = f"https://{auth.domo_instance}.domo.com/api/customer/v1/properties/groups.system.enabled"

    res = await gd.get_data(
        url=url,
        auth=auth,
        method="GET",
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Group_GET_Error(res=res)

    if res.response.get("value"):
        res.response["value"] = dmcv.convert_string_to_bool(res.response["value"])

    return res


@gd.route_function
async def toggle_system_group_visibility(
    auth,
    is_hide_system_groups: bool,
    *,
    context: RouteContext | None = None,
    debug_prn: bool = False,
    **context_kwargs,
):
    context = RouteContext.build_context(context, **context_kwargs)

    if debug_prn:
        print(
            f"toggling group visiblity in {auth.domo_instance} {'hiding system groups' if is_hide_system_groups else 'show system groups'}"
        )

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/setVisibility"

    res = await gd.get_data(
        url=url,
        method="POST",
        auth=auth,
        body={"type": "system", "hidden": is_hide_system_groups},
        context=context,
    )

    if not res.is_success:
        raise Group_CRUD_Error(res=res)

    return await is_system_groups_visible(
        auth=auth,
        context=RouteContext(
            session=context.session,
            debug_api=context.debug_api,
            debug_num_stacks_to_drop=context.debug_num_stacks_to_drop + 1,
            parent_class=context.parent_class,
        ),
    )


class GroupType_Enum(DomoEnumMixin, Enum):
    OPEN = "open"
    ADHOC = "adHoc"
    CLOSED = "closed"
    DIRECTORY = "directory"
    DYNAMIC = "dynamic"
    SYSYTEM = "system"


def generate_body_create_group(
    group_name: str, group_type: str = "open", description: str = ""
) -> dict:
    """Generates the body to create group for content_v2_group API"""
    body = {"name": group_name, "type": group_type, "description": description}

    return body


@gd.route_function
async def create_group(
    auth: DomoAuth,
    group_name: str,
    group_type: Union[GroupType_Enum, str] = "open",
    description: str = "",
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    # body : {"name": "GROUP_NAME", "type": "open", "description": ""}
    context = RouteContext.build_context(context, **context_kwargs)

    if isinstance(group_type, GroupType_Enum):
        group_type = group_type.value

    body = generate_body_create_group(
        group_name=group_name, group_type=group_type, description=description
    )
    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=body,
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        try:
            group_exists = await search_groups_by_name(
                auth=auth, search_name=group_name, is_exact_match=True
            )

            if group_exists.is_success:
                raise Group_CRUD_Error(
                    res=res,
                    message=f"{group_name} already exists. Choose a different group_name",
                )

        except SearchGroups_Error:
            raise Group_CRUD_Error(
                res=res,
                message=res.response,
            )

    return res


@gd.route_function
async def update_group(
    auth: DomoAuth,
    group_id: int,
    group_name: str = None,
    group_type: str = None,
    description: str = None,
    additional_params: dict = None,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    context = RouteContext.build_context(context, **context_kwargs)

    if isinstance(group_type, GroupType_Enum):
        group_type = group_type.value

    s = {"groupId": int(group_id)}

    if group_name:
        s.update({"name": group_name})

    if group_type:
        s.update({"type": group_type})

    if description:
        s.update({"description": description})

    if additional_params and isinstance(additional_params, dict):
        s.update({**additional_params})
        pass

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=[s],
        context=context,
    )

    if group_name and res.status == 400:
        raise Group_CRUD_Error(
            res=res,
            message="are you trying to create an account with a duplicate name?",
        )

    if not res.is_success:
        raise Group_CRUD_Error(
            res=res,
            message=res.response,
        )

    res.response = f"updated {group_id} from {auth.domo_instance}"
    return res


@gd.route_function
async def delete_groups(
    auth: DomoAuth,
    group_ids: list[str],  # list of group_ids
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    context = RouteContext.build_context(context, **context_kwargs)

    group_ids = group_ids if isinstance(group_ids, list) else [str(group_ids)]

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="DELETE",
        body=group_ids,
        context=context,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Group_CRUD_Error(
            res=res,
            message=f"failed to delete {', '.join(group_ids)}",
        )

    res.response = f"deleted {', '.join(group_ids)} from {auth.domo_instance}"
    return res


@gd.route_function
async def get_group_owners(
    auth: DomoAuth,
    group_id: str,
    *,
    context: RouteContext | None = None,
    return_raw: bool = False,
    # maximum: int = None,
    # skip: int = 0,
    # debug_loop: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    context = RouteContext.build_context(context, **context_kwargs)

    # url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/access"
    # url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/users?group={group_id}"
    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/permissions?checkOwnership=true&includeUsers=false"

    res = await gd.get_data(
        auth=auth,
        url=url,
        body=[str(group_id)],
        method="POST",
        context=context,
    )

    ## probably not paginated, headers would not make sence when checking owners

    # def arr_fn(res):
    #     return res.response[0].get("permissions").get("owners")

    # res = await gd.looper(
    #     auth=auth,
    #     session=session,
    #     url=url,
    #     offset_params={"offset": "offset", "limit": "limit"},
    #     arr_fn=arr_fn,
    #     loop_until_end=(
    #         True if maximum is None else False
    #     ),  # usually you'll set this to true.  it will override maximum
    #     method="POST",
    #     body=[str(group_id)],
    #     offset_params_in_body=False,
    #     limit=50,
    #     skip=skip,
    #     maximum=maximum,
    #     debug_api=debug_api,
    #     debug_loop=debug_loop,
    #     debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    #     parent_class=parent_class,
    #     return_raw=return_raw,
    # )

    if return_raw:
        return res

    if not res.is_success:
        raise Group_GET_Error(res=res)

    res.response = res.response[0]["permissions"]["owners"]

    return res


@gd.route_function
async def get_group_membership(
    auth: DomoAuth,
    group_id: str,
    return_raw: bool = False,
    maximum=None,
    *,
    context: RouteContext | None = None,
    debug_loop: bool = False,
    **context_kwargs,
) -> rgd.ResponseGetData:
    context = RouteContext.build_context(context, **context_kwargs)

    # url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/access"
    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/users"

    # res = await gd.get_data(
    #     auth=auth,
    #     url=url,
    #     method="GET",
    #     debug_api=debug_api,
    #     session=session,
    #     parent_class=parent_class,
    #     debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    #     params = {
    #         "offset" : 1,
    #         "limit" : 1
    #     }
    # )
    def arr_fn(res):
        return res.response.get("groupUserList")

    fixed_params = {"group": group_id}

    res = await gd.looper(
        auth=auth,
        url=url,
        fixed_params=fixed_params,
        offset_params={"offset": "offset", "limit": "limit"},
        arr_fn=arr_fn,
        loop_until_end=(
            True if maximum is None else False
        ),  # usually you'll set this to true.  it will override maximum
        maximum=maximum,
        method="GET",
        offset_params_in_body=False,
        limit=500,
        skip=0,
        context=context,
        debug_loop=debug_loop,
        return_raw=return_raw,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise Group_GET_Error(res=res)

    return res


def generate_body_update_group_membership(
    group_id: str,
    add_owner_arr: list[str] = None,
    remove_owner_arr: list[str] = None,
    remove_member_arr: list[str] = None,
    add_member_arr: list[str] = None,
) -> list[dict]:
    """
    each member or owner obj should be an object of shape {"type", "id"}
    """
    body = {"groupId": int(group_id)}

    if add_owner_arr and len(add_owner_arr) > 0:
        body.update({"addOwners": add_owner_arr})

    if remove_owner_arr and len(remove_owner_arr) > 0:
        body.update({"removeOwners": remove_owner_arr})

    if remove_member_arr and len(remove_member_arr) > 0:
        body.update({"removeMembers": remove_member_arr})
    if add_member_arr and len(add_member_arr) > 0:
        body.update({"addMembers": add_member_arr})

    return [body]


@gd.route_function
async def update_group_membership(
    auth: DomoAuth,
    group_id: str,
    update_payload: dict = None,
    add_member_arr: list[dict] = None,
    remove_member_arr: list[dict] = None,
    add_owner_arr: list[dict] = None,
    remove_owner_arr: list[dict] = None,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    each member or owner obj should be an object of shape {"type", "id"}
    """
    context = RouteContext.build_context(context, **context_kwargs)

    update_payload = update_payload or generate_body_update_group_membership(
        group_id=group_id,
        add_member_arr=add_member_arr,
        remove_member_arr=remove_member_arr,
        add_owner_arr=add_owner_arr,
        remove_owner_arr=remove_owner_arr,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/access"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=update_payload,
        context=context,
    )

    if not res.is_success:
        raise Group_CRUD_Error(res=res)

    return res


@gd.route_function
async def update_group_members(
    auth: DomoAuth,
    group_id: str,
    add_member_arr: list[dict] = None,
    remove_member_arr: list[dict] = None,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    each member or owner obj should be an object of shape {"type", "id"}
    """
    context = RouteContext.build_context(context, **context_kwargs)

    body = generate_body_update_group_membership(
        group_id=group_id,
        add_member_arr=add_member_arr,
        remove_member_arr=remove_member_arr,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/access"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        context=context,
    )

    if not res.is_success:
        raise Group_CRUD_Error(res=res)

    return res


@gd.route_function
async def update_group_owners(
    auth: DomoAuth,
    group_id: str,
    add_owner_arr: list[dict] = None,
    remove_owner_arr: list[dict] = None,
    *,
    context: RouteContext | None = None,
    **context_kwargs,
) -> rgd.ResponseGetData:
    """
    each member or owner obj should be an object of shape {"type", "id"}
    """
    context = RouteContext.build_context(context, **context_kwargs)

    body = generate_body_update_group_membership(
        group_id=group_id,
        add_owner_arr=add_owner_arr,
        remove_owner_arr=remove_owner_arr,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/groups/access"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        context=context,
    )

    if not res.is_success:
        raise Group_CRUD_Error(res=res)

    return res



__all__ = [
    "DACNoTargetInstanceError",
    "DACNoTargetUserError",
    "DACNoPasswordError",
    "DACNoUserNameError",
    "DACNoAccessTokenNameError",
    "DACNoAccessTokenError",
    "DACValidAuthError",
    "DomoAccountCredential",
]


from dataclasses import dataclass, field

import httpx

from ...auth import DomoAuth, DomoFullAuth, DomoTokenAuth
from ...base.exceptions import AuthError, ClassError, DomoError
from ...utils import convert as dmcv
from ..DomoInstanceConfig.access_token import DomoAccessToken
from ..DomoUser import DomoUser, DomoUsers
from .account_default import DomoAccount_Default


class DACNoTargetInstanceError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no target_instance on class - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DACNoTargetUserError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no target_user on class - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DACNoPasswordError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no password stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DACNoUserNameError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no username stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DACNoAccessTokenNameError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message="must pass access token name to retrieve",
            cls_instance=cls_instance,
        )


class DACNoAccessTokenError(ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no access_token stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DACValidAuthError(ClassError):
    def __init__(self, cls_instance, message=None):
        super().__init__(
            message=message
            or f"{cls_instance.name} no valid auth retrieved for domo_instance - {cls_instance.target_instance}",
            cls_instance=cls_instance,
        )


@dataclass
class DomoAccountCredential(DomoAccount_Default):
    """Account credential management class for Domo accounts.

    This class extends DomoAccount_Default to provide credential management
    capabilities including authentication testing, password management, and
    access token operations.

    Attributes:
        target_instance: Target Domo instance for credential operations
        is_valid_full_auth: Whether full authentication is valid
        is_valid_token_auth: Whether token authentication is valid
        target_auth: Active authentication object for target instance
        target_user: DomoUser object for the target user
        target_access_token: Access token for the account
    """

    target_auth: DomoAuth = field(default=None, compare=False)
    target_user: DomoUser = field(default=None, compare=False)
    target_access_token: DomoAccessToken = field(default=None, compare=False)
    target_instance: str = field(default=None)

    is_valid_full_auth: bool = None
    is_valid_token_auth: bool = None

    _token_auth: DomoAuth = field(repr=False, default=None, compare=False)
    _full_auth: DomoAuth = field(repr=False, default=None, compare=False)

    # Note: __post_init__ is inherited from DomoAccount_Default
    # which initializes the Access subentity

    @classmethod
    def from_dict(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: DomoAuth | None = None,
        is_use_default_account_class: bool = False,
        **kwargs,
    ):
        """Create Account_Credential from dictionary representation.

        Args:
            obj: Dictionary containing account data
            is_admin_summary: Whether this is an admin summary view
            auth: Authentication object
            **kwargs: Additional keyword arguments including target_instance

        Returns:
            DomoAccount_Credential instance
        """
        return super().from_dict(
            obj=obj,
            is_admin_summary=is_admin_summary,
            auth=auth,
            new_cls=cls,
            is_use_default_account_class=is_use_default_account_class,
            target_instance=kwargs.get("target_instance"),
        )

    def set_password(self, password: str) -> bool:
        """Set the password in the account configuration.

        Args:
            password: New password to set

        Returns:
            True if successful
        """
        self.Config.password = password
        return True

    def set_username(self, username: str) -> bool:
        """Set the username in the account configuration.

        Args:
            username: New username to set

        Returns:
            True if successful
        """
        self.Config.username = username
        return True

    def set_access_token(self, access_token: str) -> bool:
        """Set the access token in the account configuration.

        Args:
            access_token: New access token to set

        Returns:
            True if successful
        """
        self.Config.domo_access_token = access_token
        return True

    async def test_full_auth(
        self, debug_api: bool = False, session: httpx.AsyncClient | None = None
    ) -> bool:
        """Test full authentication (username/password) for the account.

        Generates a DomoFullAuth object and validates it against the target instance.

        Args:
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            True if authentication is valid, False otherwise

        Raises:
            DAC_NoUserName: If username is not configured
            DAC_NoPassword: If password is not configured
            DAC_NoTargetInstance: If target instance is not set
        """
        self.is_valid_full_auth = False

        if not self.Config.username:
            raise DACNoUserNameError(self)

        if not self.Config.password:
            raise DACNoPasswordError(self)

        if not self.target_instance:
            raise DACNoTargetInstanceError(self)

        self._full_auth = DomoFullAuth(
            domo_instance=self.target_instance,
            domo_username=self.Config.username,
            domo_password=self.Config.password,
        )

        try:
            await self._full_auth.print_is_token(debug_api=debug_api, session=session)
            self.is_valid_full_auth = True

        except AuthError as e:
            dmcv.print_md(f"🤯 test_full_auth for: ***{self.name}*** returned {e}")

            self.is_valid_full_auth = False

        return self.is_valid_full_auth

    async def test_token_auth(
        self, debug_api: bool = False, session: httpx.AsyncClient | None = None
    ) -> bool:
        """Test token authentication for the account.

        Generates a DomoTokenAuth object and validates it against the target instance.

        Args:
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            True if authentication is valid, False otherwise

        Raises:
            DAC_NoAccessToken: If access token is not configured
            DAC_NoTargetInstance: If target instance is not set
        """

        self.is_valid_token_auth = False

        if not self.Config.domo_access_token:
            raise DACNoAccessTokenError(self)

        if not self.target_instance:
            raise DACNoTargetInstanceError(self)

        self._token_auth = DomoTokenAuth(
            domo_instance=self.target_instance,
            domo_access_token=self.Config.domo_access_token,
        )

        try:
            await self._token_auth.print_is_token(debug_api=debug_api, session=session)
            self.is_valid_token_auth = True
            self.target_auth = self._token_auth

        except AuthError as e:
            dmcv.print_md(f"🤯 test_token_auth for: ***{self.name}*** returned {e}")
            self.is_valid_token_auth = False

        return self.is_valid_token_auth

    def _set_target_auth(
        self,
        valid_backup_auth: DomoAuth | None = None,
    ) -> DomoAuth:
        """Set target authentication using best available method.

        Prioritizes authentication methods in order: Token Auth > Full Auth > Backup Auth

        Args:
            valid_backup_auth: Validated backup authentication object to use as fallback

        Returns:
            The selected authentication object

        Raises:
            DACValidAuthError: If no valid authentication method is available
        """

        target_auth = None

        if self.is_valid_token_auth:
            target_auth = self._token_auth

        if self.is_valid_full_auth:
            target_auth = self._full_auth

        if not target_auth and (valid_backup_auth and valid_backup_auth.is_valid_token):
            target_auth = valid_backup_auth

        if not target_auth:
            raise DACValidAuthError(self)

        self.target_auth = target_auth

        return self.target_auth

    async def test_auths(
        self,
        backup_auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> dict:
        """Test both token and full authentication methods.

        Attempts to validate both authentication methods and sets the best available
        as the target authentication.

        Args:
            backup_auth: Backup authentication to use if configured auths fail
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            Dictionary with authentication test results
        """
        ## test token auth
        try:
            await self.test_token_auth(debug_api=debug_api, session=session)

        except DomoError as e:
            print(f"testing token: {self.name}: {e}")

        ## test full auth
        try:
            await self.test_full_auth(debug_api=debug_api, session=session)

        except DomoError as e:
            print(f"testing full auth: {self.name}: {e}")

        ## generate target_auth
        try:
            self._set_target_auth(valid_backup_auth=backup_auth)
        except DACValidAuthError as e:
            print(f"{self.name}: unable to generate valid target_auth: {e}")
        return self.to_dict()

    def to_dict(self, return_snake_case: bool = False) -> dict:
        """Convert credential information to dictionary.

        Returns:
            Dictionary containing account ID, alias, instance, and auth validity status
        """

        s = super().to_dict(return_snake_case=return_snake_case)
        s.update(
            {
                "account_id": self.id,
                "alias": self.name,
                "target_instance": self.target_instance,
                "is_valid_full_auth": self.is_valid_full_auth,
                "is_valid_token_auth": self.is_valid_token_auth,
            }
        )
        return s

    async def get_target_user(
        self,
        user_email: str | None = None,
        target_auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> DomoUser:
        """Retrieve the target user for this account.

        Args:
            user_email: Email address of user (defaults to configured username)
            target_auth: Authentication object (defaults to self.target_auth)
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            DomoUser object for the target user

        Raises:
            DAC_NoUserName: If user email is not provided or configured
            DACValidAuthError: If target authentication is not available
            DAC_NoTargetUser: If user cannot be found
        """
        user_email = user_email or self.Config.username

        if not user_email:
            raise DACNoUserNameError(self)

        target_auth = target_auth or self.target_auth

        if not target_auth:
            raise DACValidAuthError(
                self,
                message="no target_auth, pass a valid backup_auth",
            )

        self.target_user = await DomoUsers(auth=target_auth).search_by_email(
            email=[user_email],
            debug_api=debug_api,
            session=session,
        )

        if not self.target_user:
            raise DACNoTargetUserError(self)

        return self.target_user

    async def update_target_user_password(
        self,
        new_password: str,
        user_email: str | None = None,
        is_update_account: bool = True,
        target_auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> "DomoAccountCredential":
        """Update the password for the target user.

        Args:
            new_password: New password to set for the user
            user_email: Email address of user (defaults to configured username)
            is_update_account: Whether to update the account config with new password
            target_auth: Authentication object (defaults to self.target_auth)
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            Self for method chaining

        Raises:
            DACValidAuthError: If target authentication is not available
        """
        target_auth = target_auth or self.target_auth

        if not target_auth:
            raise DACValidAuthError(
                self,
                message="no target_auth, pass a valid backup_auth",
            )

        if not self.target_user:
            await self.get_target_user(
                debug_api=debug_api,
                session=session,
                user_email=user_email,
                target_auth=target_auth,
            )

        await self.target_user.reset_password(
            new_password=new_password,
            debug_api=debug_api,
            session=session,
        )

        self.set_password(new_password)

        if is_update_account:
            await self.update_config(debug_api=debug_api, session=session)

        return self

    async def get_target_access_token(
        self,
        token_name: str | None = None,
        user_email: str | None = None,
        target_auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> DomoAccessToken:
        """Retrieve an access token for the target user.

        Args:
            token_name: Name of the access token (defaults to account name)
            user_email: Email address of user (defaults to configured username)
            target_auth: Authentication object (defaults to self.target_auth)
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            DomoAccessToken object if found, None otherwise

        Raises:
            DACValidAuthError: If target authentication is not available
            DAC_NoAccessTokenName: If token name is not provided or available
        """
        target_auth = target_auth or self.target_auth

        if not target_auth:
            raise DACValidAuthError(
                self,
                message="no target_auth, pass a valid backup_auth",
            )

        if not self.target_user:
            await self.get_target_user(
                debug_api=debug_api,
                session=session,
                user_email=user_email,
                target_auth=target_auth,
            )

        token_name = token_name or self.name

        if not token_name:
            raise DACNoAccessTokenNameError(self)

        domo_access_tokens = await self.target_user.get_access_tokens(
            session=session,
            debug_api=debug_api,
        )

        self.target_access_token = next(
            (
                dat
                for dat in domo_access_tokens
                if dat and (dat.name and dat.name.lower() == token_name.lower())
            ),
            None,
        )

        if not self.target_access_token:
            raise DACNoAccessTokenNameError(self)

        return self.target_access_token

    async def regenerate_target_access_token(
        self,
        token_name: str | None = None,
        duration_in_days: int = 90,
        user_email: str | None = None,
        is_update_account: bool = True,
        target_auth: DomoAuth | None = None,
        debug_api: bool = False,
        session: httpx.AsyncClient | None = None,
    ) -> "DomoAccountCredential":
        """Regenerate or create an access token for the target user.

        If a token with the given name exists, it will be regenerated. Otherwise,
        a new token will be created.

        Args:
            token_name: Name of the access token (defaults to account name)
            duration_in_days: Token validity duration in days (default: 90)
            user_email: Email address of user (defaults to configured username)
            is_update_account: Whether to update the account config with new token
            target_auth: Authentication object (defaults to self.target_auth)
            debug_api: Enable API debugging
            session: HTTP client session (optional)

        Returns:
            Self for method chaining

        Raises:
            DACValidAuthError: If target authentication is not available
            DAC_NoTargetUser: If target user cannot be retrieved
        """
        target_auth = target_auth or self.target_auth

        if not target_auth:
            raise DACValidAuthError(
                self,
                message="no target_auth, pass a valid backup_auth",
            )

        domo_access_token = await self.get_target_access_token(
            token_name=token_name,
            user_email=user_email,
            target_auth=target_auth,
            debug_api=debug_api,
            session=session,
        )  # handles retrieving target user

        if not self.target_user:
            raise DACNoTargetUserError(self)

        if domo_access_token:
            await domo_access_token.regenerate(
                duration_in_days=duration_in_days, session=session, debug_api=debug_api
            )

        else:
            domo_access_token = await DomoAccessToken.generate(
                duration_in_days=duration_in_days,
                token_name=token_name,
                auth=target_auth,
                owner=self.target_user,
                debug_api=debug_api,
                session=session,
            )

            self.target_access_token = domo_access_token

        self.set_access_token(domo_access_token.token)

        if is_update_account:
            await self.update_config(debug_api=debug_api, session=session)

        return self



__all__ = ["DomoAccount", "DomoAccounts_NoAccount", "DomoAccounts"]


from dataclasses import dataclass

import httpx

from ...auth import DomoAuth
from ...base import exceptions as dmde
from ...base.entities import DomoManager
from ...routes import (
    account as account_routes,
    datacenter as datacenter_routes,
)
from ...utils import chunk_execution as dmce
from .account_credential import DomoAccountCredential
from .account_default import (
    DomoAccount_Default,
    UpsertAccount_MatchCriteriaError,
)
from .account_oauth import DomoAccount_OAuth
from .config import AccountConfig


class DomoAccounts_NoAccount(dmde.ClassError):
    def __init__(self, cls=None, cls_instance=None, message=None, domo_instance=None):
        super().__init__(
            cls=cls, cls_instance=cls_instance, message=message, entity_id=domo_instance
        )


@dataclass
class DomoAccount(DomoAccount_Default):
    @classmethod
    def from_dict(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: DomoAuth = None,
        is_use_default_class=False,
        **kwargs,
    ):
        """converts data_v1_accounts API response into an accounts class object"""

        if is_use_default_class:
            new_cls = cls

        if obj.get("credentialsType") == "oauth":
            new_cls = DomoAccount_OAuth
        else:
            new_cls = DomoAccountCredential

        return super().from_dict(
            auth=auth,
            obj=obj,
            is_admin_summary=is_admin_summary,
            **{**kwargs, "new_cls": new_cls},
        )


@dataclass
class DomoAccounts(DomoManager):
    accounts: list[DomoAccount] = None
    oauths: list[DomoAccount_OAuth] = None

    async def get_accounts_accountsapi(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        is_use_default_account_class: bool = True,
        debug_num_stacks_to_drop: int = 2,
    ):
        res = await account_routes.get_accounts(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop - 1,
        )

        if return_raw:
            return res

        if len(res.response) == 0:
            self.accounts = []
            return self.accounts

        self.accounts = await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoAccount.get_by_id(
                    account_id=obj.get("id"),
                    debug_api=debug_api,
                    session=session,
                    auth=self.auth,
                    is_use_default_account_class=is_use_default_account_class,
                )
                for obj in res.response
            ],
        )

        return self.accounts

    async def get_accounts_queryapi(
        self,
        additional_filters_ls=None,
        is_use_default_account_class: bool = False,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):
        """v2 api for works with group_account_v2 beta"""

        from ...routes import datacenter as datacenter_routes

        res = await datacenter_routes.search_datacenter(
            auth=self.auth,
            entity_type=datacenter_routes.Datacenter_Enum.ACCOUNT.value,
            additional_filters_ls=additional_filters_ls,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop - 1,
        )

        if return_raw:
            return res

        if len(res.response) == 0:
            self.accounts = []
            return self.accounts

        self.accounts = [
            DomoAccount.from_dict(
                account_obj,
                auth=self.auth,
                is_use_default_class=is_use_default_account_class,
            )
            for account_obj in res.response
        ]
        return self.accounts

    async def get(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        is_use_default_account_class: bool = False,
        debug_num_stacks_to_drop: int = 3,
    ):
        domo_accounts = None
        try:
            domo_accounts = await self.get_accounts_queryapi(
                debug_api=debug_api,
                session=session,
                is_use_default_account_class=is_use_default_account_class,
                return_raw=return_raw,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop - 1,
            )

        except datacenter_routes.SearchDatacenter_NoResultsFound as e:
            print(e)

        if not domo_accounts:
            domo_accounts = await self.get_accounts_accountsapi(
                debug_api=debug_api,
                session=session,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop - 1,
                is_use_default_account_class=is_use_default_account_class,
                return_raw=return_raw,
            )

        if return_raw:
            return domo_accounts

        return self.accounts

    async def get_oauths(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop: int = 2,
    ):
        res = await account_routes.get_oauth_accounts(
            auth=self.auth,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

        if return_raw:
            return res

        self.oauths = [
            DomoAccount_OAuth.from_dict(
                obj=obj, auth=self.auth, is_use_default_account_class=True
            )
            for obj in res.response
        ]

        return self.oauths

    async def search_by_name(
        self,
        account_name: str,
        data_provider_type: str = None,
        is_use_default_account_class: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        is_suppress_not_found_exception: bool = False,
        **kwargs,
    ) -> DomoAccount | None:
        """Search for an account by name (matches display_name or name).

        Args:
            account_name: Account name to search for (case-insensitive)
            data_provider_type: Optional filter by data provider type
            is_use_default_account_class: Use default account class
            debug_api: Enable debug output
            session: Optional httpx session

        Returns:
            Matching DomoAccount or None if not found

        Raises:
            DomoError: If account retrieval fails
        """
        await self.get(
            debug_api=debug_api,
            session=session,
            is_use_default_account_class=is_use_default_account_class,
            **kwargs,
        )

        for account in self.accounts:
            # Check both name and display_name for matching
            name_match = False
            if (
                account.display_name
                and account.display_name.lower() == account_name.lower()
            ):
                name_match = True
            elif account.name and account.name.lower() == account_name.lower():
                name_match = True

            if not name_match:
                continue

            # If data_provider_type specified, must match
            if data_provider_type and data_provider_type != account.data_provider_type:
                continue

            return account

        if is_suppress_not_found_exception:
            return None

        raise DomoAccounts_NoAccount(
            cls=self.__class__,
            message=f"No account found with name '{account_name}'",
            domo_instance=self.auth.domo_instance,
        )

    @classmethod
    async def upsert_account(
        cls,
        auth: DomoAuth,
        account_id: str = None,
        account_name: str = None,
        account_config: AccountConfig = None,
        data_provider_type: str = None,
        debug_api: bool = False,
        debug_prn: bool = False,
        return_raw: bool = False,
        return_search: bool = False,
        is_use_default_account_class: bool = True,
        session: httpx.AsyncClient = None,
        **kwargs,
    ):
        """search for an account and upsert it"""

        if not account_name and not account_id:
            raise UpsertAccount_MatchCriteriaError(domo_instance=auth.domo_instance)

        data_provider_type = (
            data_provider_type or account_config and account_config.data_provider_type
        )
        acc = None

        if account_id:
            try:
                acc = await DomoAccount.get_by_id(
                    auth=auth,
                    session=session,
                    debug_api=debug_api,
                    account_id=account_id,
                    is_use_default_account_class=is_use_default_account_class,
                    **kwargs,
                )
            except dmde.DomoError:
                pass

        if account_name and not acc:
            try:
                domo_accounts = DomoAccounts(auth=auth)
                acc = await domo_accounts.search_by_name(
                    account_name=account_name,
                    data_provider_type=data_provider_type,
                    is_use_default_account_class=is_use_default_account_class,
                    debug_api=debug_api,
                    session=session,
                    **kwargs,
                )
            except dmde.DomoError:
                pass

        if return_search:
            return acc

        if not isinstance(
            acc, (DomoAccount_Default, DomoAccount, DomoAccountCredential)
        ):
            if debug_prn:
                print(f"creating {account_name} in {auth.domo_instance}")

            return await DomoAccount.create_account(
                account_name=account_name,
                config=account_config,
                auth=auth,
                debug_api=debug_api,
                return_raw=return_raw,
            )

        if account_name and account_id:
            if debug_prn:
                print(
                    f"upsert-ing {acc.id} - {acc.display_name or acc.name} in {auth.domo_instance}"
                )

            await acc.update_name(
                account_name=account_name,
                debug_api=debug_api,
                return_raw=return_raw,
                session=session,
            )

        if account_config:  # upsert account
            acc.Config = account_config

            if debug_prn:
                print(f"upsertting {acc.id}:  updating config")

            await acc.update_config(
                debug_api=debug_api, return_raw=return_raw, session=session
            )

        return acc



"""Page access control and sharing operations."""

__all__ = ["test_page_access", "get_accesslist", "share"]

from typing import Union

import httpx

from ...auth import DomoAuth
from ...client.response import ResponseGetData
from ...routes import (
    datacenter as datacenter_routes,
    page as page_routes,
)
from ...utils import chunk_execution as dmce
from .. import DomoUser as dmu
from .exceptions import Page_NoAccess


async def test_page_access(
    self,
    suppress_no_access_error: bool = False,
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_num_stacks_to_drop: int = 2,
) -> ResponseGetData:
    """Test if the authenticated user has access to the page.

    This method calls the page access test API endpoint which returns the page owners.
    If the user doesn't have access, it raises a Page_NoAccess exception unless suppressed.

    Args:
        suppress_no_access_error: If True, suppresses the Page_NoAccess exception when
            user doesn't have access. Defaults to False.
        debug_api: Enable detailed API request/response logging. Defaults to False.
        return_raw: Return raw ResponseGetData without processing. Defaults to False.
        session: Optional httpx client session for connection reuse.
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output. Defaults to 2.

    Returns:
        ResponseGetData object containing page access information and owner list.

    Raises:
        Page_NoAccess: If user doesn't have access and suppress_no_access_error is False.
    """

    res = await page_routes.get_page_access_test(
        auth=self.auth,
        page_id=self.id,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    try:
        page_access = res.response.get("pageAccess")

        if not page_access:
            raise Page_NoAccess(
                page_id=self.id,
                page_title=self.title,
                domo_instance=self.auth.domo_instance,
                function_name=res.traceback_details.function_name,
                parent_class=self.__class__.__name__,
            )

    except Page_NoAccess as e:
        print(e)

        if not suppress_no_access_error:
            raise e

    return res


async def get_accesslist(
    self,
    auth: DomoAuth | None = None,
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_num_stacks_to_drop: int = 2,
) -> Union[ResponseGetData, dict[str, Union[int, list]]]:
    """Retrieve the access list for the page showing users and groups with access.

    This method fetches the comprehensive access list for a page, including:
    - Users with explicit share permissions
    - Groups with access permissions
    - User ownership information (direct or through group membership)
    - Group ownership information

    Args:
        auth: Authentication object. If not provided, uses self.auth.
        return_raw: Return raw ResponseGetData without processing. Defaults to False.
        debug_api: Enable detailed API request/response logging. Defaults to False.
        session: Optional httpx client session for connection reuse.
        debug_num_stacks_to_drop: Number of stack frames to drop in debug output. Defaults to 2.

    Returns:
        If return_raw is True:
            ResponseGetData object containing raw access list data.
        If return_raw is False:
            Dictionary containing:
                - explicit_shared_user_count (int): Number of users with explicit shares
                - total_user_count (int): Total number of users with access
                - domo_users (list[DomoUser]): list of users with access, enriched with:
                    - custom_attributes['is_explicit_share']: True if directly shared
                    - custom_attributes['group_membership']: list of groups user belongs to
                    - custom_attributes['is_owner']: True if user is an owner
                - domo_groups (list[DomoGroup]): list of groups with access, enriched with:
                    - custom_attributes['is_owner']: True if group is an owner

    Raises:
        PageSharing_Error: If access list retrieval fails (raised by route function).
        SearchPageNotFoundError: If page with specified ID doesn't exist (raised by route function).
    """
    auth = auth or self.auth

    res = await page_routes.get_page_access_list(
        auth=auth,
        is_expand_users=True,
        page_id=self.id,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    from ..DomoGroup import core as dmg

    s = {
        "explicit_shared_user_count": res.response.get("explicitSharedUserCount"),
        "total_user_count": res.response.get("totalUserCount"),
    }

    user_ls = res.response.get("users", None)
    domo_users = []

    if user_ls and isinstance(user_ls, list) and len(user_ls) > 0:
        domo_users = await dmu.DomoUsers.by_id(
            user_ids=[user.get("id") for user in user_ls],
            only_allow_one=False,
            auth=auth,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

    group_ls = res.response.get("groups", None)
    domo_groups = []
    if group_ls and isinstance(group_ls, list) and len(group_ls) > 0:
        domo_groups = await dmce.gather_with_concurrency(
            n=60,
            *[
                dmg.DomoGroup.get_by_id(
                    group_id=group.get("id"),
                    auth=auth,
                    session=session,
                    debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
                )
                for group in group_ls
            ],
        )

    res = await self.test_page_access(suppress_no_access_error=True)
    owner_ls = res.response["owners"]  # from test_page_access

    for domo_user in domo_users:
        # isExplicitShare is set by the get_access_list API response
        domo_user.custom_attributes["is_explicit_share"] = next(
            obj["isExplicitShare"]
            for obj in user_ls
            if int(obj.get("id")) == int(domo_user.id)
        )

        # group membership is determined by get_access_list API response
        domo_user.custom_attributes["group_membership"] = [
            domo_group
            for group_obj in group_ls
            for domo_group in domo_groups
            if int(domo_user.id) in [int(obj["id"]) for obj in group_obj.get("users")]
            and domo_group.id == group_obj["id"]
        ]

        # isOwner determined by test_access API response and group membership
        domo_user.custom_attributes["is_owner"] = False

        # test ownership as a user
        match_owner = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"]) == int(domo_user.id)
                and owner_obj["type"] == "USER"
            ),
            None,
        )

        match_group = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"])
                in [
                    int(domo_group.id)
                    for domo_group in domo_user.custom_attributes["group_membership"]
                ]
                and owner_obj["type"] == "GROUP"
            ),
            None,
        )

        if match_owner or match_group:
            domo_user.custom_attributes["is_owner"] = True

    # group ownership is confirmed test_access API
    for domo_group in domo_groups:
        match_owner = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"]) == int(domo_group.id)
                and owner_obj["type"] == "GROUP"
            ),
            None,
        )

        domo_group.custom_attributes["is_owner"] = True if match_owner else False

    return {
        **s,
        "domo_users": domo_users,
        "domo_groups": domo_groups,
    }


async def share(
    self,
    auth: DomoAuth | None = None,
    domo_users: Union[list, object] | None = None,
    domo_groups: Union[list, object] | None = None,
    message: str | None = None,
    debug_api: bool = False,
    session: httpx.AsyncClient | None = None,
) -> ResponseGetData:
    """Share the page with specified users and/or groups.

    This method shares the page with one or more users and/or groups. It uses the
    datacenter share_resource route function to perform the sharing operation.

    Args:
        auth: Authentication object. If not provided, uses self.auth.
        domo_users: DomoUser object(s) to share page with. Can be a single DomoUser
            or a list of DomoUser objects. Defaults to None.
        domo_groups: DomoGroup object(s) to share page with. Can be a single DomoGroup
            or a list of DomoGroup objects. Defaults to None.
        message: Optional message to include in the automated email notification.
        debug_api: Enable detailed API request/response logging. Defaults to False.
        session: Optional httpx client session for connection reuse.

    Returns:
        ResponseGetData object containing the sharing operation result.

    Raises:
        Exception: Various exceptions may be raised by the datacenter share_resource
            route function if the sharing operation fails.
    """
    if domo_groups:
        domo_groups = domo_groups if isinstance(domo_groups, list) else [domo_groups]
    if domo_users:
        domo_users = domo_users if isinstance(domo_users, list) else [domo_users]

    res = await datacenter_routes.share_resource(
        auth=auth or self.auth,
        resource_ids=[self.id],
        resource_type=datacenter_routes.ShareResource_Enum.PAGE,
        group_ids=[group.id for group in domo_groups] if domo_groups else None,
        user_ids=[user.id for user in domo_users] if domo_users else None,
        message=message,
        debug_api=debug_api,
        session=session,
    )

    return res



"""Stream configuration mappings organized by platform.

This package contains stream configuration classes organized by major platform
(Snowflake, AWS, Domo, Google, etc.).

TWO PATTERNS AVAILABLE:

1. NEW PATTERN (Recommended): Typed Config Classes
   - Follows AccountConfig pattern
   - Type-safe attribute access: config.query
   - IDE autocomplete works
   - Single config object per stream

   Example:
       from domolibrary2.classes.DomoDataset.stream_configs import StreamConfig_Base
       from domolibrary2.classes.DomoDataset.stream_configs.snowflake import Snowflake_StreamConfig

       config = Snowflake_StreamConfig.from_dict({"query": "SELECT *", "databaseName": "SA_PRD"})
       query = config.query  # Type-safe, autocomplete works!

2. OLD PATTERN (Deprecated): Mapping Classes
   - Maps field names to config parameters
   - Search through list of StreamConfig objects
   - Kept for backward compatibility

   Example:
       from domolibrary2.classes.DomoDataset.stream_configs import StreamConfig_Mapping

       mapping = StreamConfig_Mappings('snowflake')
       sql_param = mapping.value.sql  # "query"
"""

# Import and re-export base classes and utilities
from ._base import (
    _MAPPING_REGISTRY,
    Stream_CRUD_Error,
    Stream_GET_Error,
    StreamConfig,
    StreamConfig_Base,
    StreamConfig_Mapping,
    StreamConfig_Mappings,
    register_mapping,
    register_stream_config,
)
from ._conformed import CONFORMED_PROPERTIES, ConformedProperty

# Import new typed config classes
from ._default import Default_StreamConfig
from ._repr import (
    ConformedPropertyReprMixin,
    create_stream_repr,
    get_available_config_keys,
    get_conformed_properties_for_repr,
    get_missing_mappings,
)
from .aws import (
    AmazonAthenaHighBandwidth_StreamConfig,
    AmazonS3AssumeRole_StreamConfig,
    AWSAthena_StreamConfig,
)
from .domo import DatasetCopy_StreamConfig, DomoCSV_StreamConfig
from .google import GoogleSheets_StreamConfig, GoogleSpreadsheets_StreamConfig
from .other import (
    AdobeAnalyticsV2_StreamConfig,
    PostgreSQL_StreamConfig,
    Qualtrics_StreamConfig,
    SharePointOnline_StreamConfig,
)
from .snowflake import (
    Snowflake_StreamConfig,
    SnowflakeKeyPairAuth_StreamConfig,
)

__all__ = [
    # NEW PATTERN (Recommended)
    "StreamConfig_Base",
    "register_stream_config",
    # Conformed properties (semantic layer)
    "ConformedProperty",
    "CONFORMED_PROPERTIES",
    # Custom repr utilities
    "create_stream_repr",
    "get_conformed_properties_for_repr",
    "get_missing_mappings",
    "get_available_config_keys",
    "ConformedPropertyReprMixin",
    # NEW: Snowflake configs
    "Snowflake_StreamConfig",
    "SnowflakeKeyPairAuth_StreamConfig",
    # NEW: AWS configs
    "AWSAthena_StreamConfig",
    "AmazonAthenaHighBandwidth_StreamConfig",
    "AmazonS3AssumeRole_StreamConfig",
    # NEW: Domo configs
    "DatasetCopy_StreamConfig",
    "DomoCSV_StreamConfig",
    # NEW: Google configs
    "GoogleSheets_StreamConfig",
    "GoogleSpreadsheets_StreamConfig",
    # NEW: Other configs
    "AdobeAnalyticsV2_StreamConfig",
    "PostgreSQL_StreamConfig",
    "Qualtrics_StreamConfig",
    "SharePointOnline_StreamConfig",
    # NEW: Default config
    "Default_StreamConfig",
    # OLD PATTERN (Deprecated but kept for compatibility)
    "StreamConfig_Mapping",
    "StreamConfig_Mappings",
    "StreamConfig",
    "_MAPPING_REGISTRY",
    "register_mapping",
    # Route exceptions
    "Stream_GET_Error",
    "Stream_CRUD_Error",
]



__all__ = [
    "AppStudio_GET_Error",
    "AppStudio_CRUD_Error",
    "AppStudioSharing_Error",
    "get_appstudio_by_id",
    "get_appstudio_access",
    "get_appstudios_adminsummary",
    "generate_body_add_page_owner_appstudios",
    "generate_body_share_appstudio",
    "add_page_owner",
    "share",
]


import httpx

from ..auth import DomoAuth
from ..base import exceptions as dmde
from ..client import (
    get_data as gd,
    response as rgd,
)


class AppStudio_GET_Error(dmde.RouteError):
    """Raised when AppStudio retrieval operations fail."""

    def __init__(
        self,
        appstudio_id: str | None = None,
        res: rgd.ResponseGetData | None = None,
        message: str | None = None,
        **kwargs,
    ):
        super().__init__(
            message=message or "AppStudio retrieval failed",
            entity_id=appstudio_id,
            res=res,
            **kwargs,
        )


class AppStudio_CRUD_Error(dmde.RouteError):
    """Raised when AppStudio create, update, or delete operations fail."""

    def __init__(
        self,
        operation: str = "operation",
        appstudio_id: str | None = None,
        res: rgd.ResponseGetData | None = None,
        message: str | None = None,
        **kwargs,
    ):
        super().__init__(
            message=message or f"AppStudio {operation} operation failed",
            entity_id=appstudio_id,
            res=res,
            **kwargs,
        )


class AppStudioSharing_Error(dmde.RouteError):
    """Raised when AppStudio sharing operations fail."""

    def __init__(
        self,
        appstudio_id: str | None = None,
        res: rgd.ResponseGetData | None = None,
        message: str | None = None,
        **kwargs,
    ):
        super().__init__(
            message=message or "AppStudio sharing operation failed",
            entity_id=appstudio_id,
            res=res,
            **kwargs,
        )


@gd.route_function
async def get_appstudio_by_id(
    auth: DomoAuth,
    appstudio_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieves an AppStudio page by ID.

    Args:
        auth: Authentication object
        appstudio_id: AppStudio identifier
        session: Optional HTTP session
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Number of stack frames to drop in traceback
        parent_class: Name of the calling class
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object

    Raises:
        AppStudio_GET_Error: If retrieval fails
    """
    # 9/21/2023 - the domo UI uses /cards to get page info
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/dataapps/{appstudio_id}?authoring=true&includeHiddenViews=true"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppStudio_GET_Error(
            appstudio_id=appstudio_id,
            res=res,
        )

    return res


@gd.route_function
async def get_appstudio_access(
    auth: DomoAuth,
    appstudio_id: str,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieves access list for an AppStudio page.

    Args:
        auth: Authentication object
        appstudio_id: AppStudio identifier
        session: Optional HTTP session
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Number of stack frames to drop in traceback
        parent_class: Name of the calling class
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing users and groups the page is shared with

    Raises:
        AppStudio_GET_Error: If retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/dataapps/{appstudio_id}/access"

    res = await gd.get_data(
        url,
        method="GET",
        auth=auth,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppStudio_GET_Error(
            appstudio_id=appstudio_id,
            res=res,
        )

    return res


@gd.route_function
async def get_appstudios_adminsummary(
    auth: DomoAuth,
    limit: int = 35,
    session: httpx.AsyncClient | None = None,
    debug_loop: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Retrieves all AppStudio pages in instance user is able to see.

    Args:
        auth: Authentication object
        limit: Maximum number of items per page
        session: Optional HTTP session
        debug_loop: Enable loop debugging
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Number of stack frames to drop in traceback
        parent_class: Name of the calling class
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object containing all accessible AppStudio pages

    Raises:
        AppStudio_GET_Error: If retrieval fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/dataapps/adminsummary"

    offset_params = {
        "offset": "skip",
        "limit": "limit",
    }

    body = {"orderBy": "title", "ascending": True}

    def arr_fn(res):
        return res.response.get("dataAppAdminSummaries")

    res = await gd.looper(
        auth=auth,
        method="POST",
        url=url,
        arr_fn=arr_fn,
        offset_params=offset_params,
        session=session,
        loop_until_end=True,
        body=body,
        limit=limit,
        debug_loop=debug_loop,
        debug_api=debug_api,
        parent_class=parent_class,
        return_raw=return_raw,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppStudio_GET_Error(res=res)

    return res


def generate_body_add_page_owner_appstudios(
    appstudio_id_ls: list[int],
    group_id_ls: list[int] | None = None,
    user_id_ls: list[int] | None = None,
    note: str = "",
    send_email: bool = False,
) -> dict:
    """Generates request body for adding page owners to AppStudio pages.

    Args:
        appstudio_id_ls: list of AppStudio IDs
        group_id_ls: Optional list of group IDs
        user_id_ls: Optional list of user IDs
        note: Optional note to include
        send_email: Whether to send email notifications

    Returns:
        Dictionary containing the request body
    """
    group_id_ls = group_id_ls or []
    user_id_ls = user_id_ls or []
    owners = []

    for group in group_id_ls:
        owners.append({"id": group, "type": "GROUP"})
    for user in user_id_ls:
        owners.append({"id": user, "type": "USER"})

    body = {
        "entityIds": appstudio_id_ls,
        "owners": owners,
        "note": note,
        "sendEmail": send_email,
    }

    return body


def generate_body_share_appstudio(
    appstudio_ids: list[int],
    group_ids: list[int] | None = None,
    user_ids: list[int] | None = None,
    message: str | None = None,
) -> dict:
    """Generates request body for sharing AppStudio pages.

    Args:
        appstudio_ids: list of AppStudio IDs to share
        group_ids: Optional list of group IDs
        user_ids: Optional list of user IDs
        message: Optional message to include

    Returns:
        Dictionary containing the request body
    """
    group_ids = group_ids or []
    user_ids = user_ids or []

    appstudio_ids = (
        appstudio_ids if isinstance(appstudio_ids, list) else [appstudio_ids]
    )
    if group_ids:
        group_ids = (
            group_ids and group_ids if isinstance(group_ids, list) else [group_ids]
        )

    if user_ids:
        user_ids = user_ids if isinstance(user_ids, list) else [user_ids]

    recipient_ls = []

    if group_ids:
        for gid in group_ids:
            recipient_ls.append({"type": "group", "id": str(gid)})

    if user_ids:
        for uid in user_ids:
            recipient_ls.append({"type": "user", "id": str(uid)})

    body = {
        "dataAppIds": appstudio_ids,
        "recipients": recipient_ls,
        "message": message,
    }

    return body


@gd.route_function
async def add_page_owner(
    auth: DomoAuth,
    appstudio_id_ls: list[int],
    group_id_ls: list[int] | None = None,
    user_id_ls: list[int] | None = None,
    note: str = "",
    send_email: bool = False,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Adds page owners to AppStudio pages.

    Args:
        auth: Authentication object
        appstudio_id_ls: list of AppStudio IDs
        group_id_ls: Optional list of group IDs to add as owners
        user_id_ls: Optional list of user IDs to add as owners
        note: Optional note to include with the change
        send_email: Whether to send email notifications
        session: Optional HTTP session
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Number of stack frames to drop in traceback
        parent_class: Name of the calling class
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object

    Raises:
        AppStudioSharing_Error: If sharing operation fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/dataapps/bulk/owners"

    body = generate_body_add_page_owner_appstudios(
        appstudio_id_ls=appstudio_id_ls,
        group_id_ls=group_id_ls,
        user_id_ls=user_id_ls,
        note=note,
        send_email=send_email,
    )

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppStudioSharing_Error(res=res)

    res.response = f"{appstudio_id_ls} appstudios successfully shared with {', '.join([recipient['id'] for recipient in body['owners']])} as owners"

    return res


@gd.route_function
async def share(
    auth: DomoAuth,
    appstudio_ids: list[int],
    group_ids: list[int] | None = None,
    user_ids: list[int] | None = None,
    message: str | None = None,
    session: httpx.AsyncClient | None = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str | None = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """Shares AppStudio pages with users or groups.

    Args:
        auth: Authentication object
        appstudio_ids: list of AppStudio IDs to share
        group_ids: Optional list of group IDs to share with
        user_ids: Optional list of user IDs to share with
        message: Optional message to include in email notification
        session: Optional HTTP session
        debug_api: Enable API debugging
        debug_num_stacks_to_drop: Number of stack frames to drop in traceback
        parent_class: Name of the calling class
        return_raw: Return raw response without processing

    Returns:
        ResponseGetData object

    Raises:
        AppStudioSharing_Error: If sharing operation fails
    """
    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/dataapps/share?sendEmail=false"

    body = generate_body_share_appstudio(
        appstudio_ids=appstudio_ids,
        group_ids=group_ids,
        user_ids=user_ids,
        message=message,
    )

    res = await gd.get_data(
        url,
        method="POST",
        auth=auth,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppStudioSharing_Error(res=res)

    res.response = f"{appstudio_ids} appstudios successfully shared with {', '.join([recipient['id'] for recipient in body['recipients']])}"

    return res
